"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const fs = require("fs");
const path = require("path");
class ScriptBuilder {
    constructor(_extratermCookieName, _extratermCookieValue) {
        this._extratermCookieName = _extratermCookieName;
        this._extratermCookieValue = _extratermCookieValue;
    }
    build() {
        return [{ type: 'text', text: this._buildCookie() + '\n' + this._buildCommands() + '\n' + this._buildShellReporting() + '\n' }];
    }
    _buildCommands() {
        return this._buildScriptForCommand('from', 'exfrom') + this._buildScriptForCommand('show', 'exshow');
    }
    _buildScriptForCommand(commandName, commandPyFile) {
        const flattener = new PythonFileFlattener(path.join(__dirname, '../../../src/commands/'));
        const script = flattener.readAndInlineCommand(commandPyFile);
        return this._formatCommand(commandName, script);
    }
    _escapeShellChars(source) {
        return source.replace(/'/g, "'\\''");
    }
}
const CTRL_C = '\x03';
const EOT = '\x04';
class FishScriptBuilder extends ScriptBuilder {
    build() {
        return [
            { type: 'text', text: CTRL_C },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty -echo; source\n' },
            { type: 'wait', durationMilliseconds: 300 },
            ...super.build(),
            { type: 'text', text: 'echo "Shell integration is configured"\n' },
            { type: 'text', text: EOT },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty echo\n' },
        ];
    }
    _buildCookie() {
        return `set -x ${this._extratermCookieName} ${this._extratermCookieValue}`;
    }
    _buildShellReporting() {
        return `
function extraterm_preexec -e fish_preexec
  echo -n -e -s "\\033&" $${this._extratermCookieName} ";2;fish\\007"
  echo -n $argv[1]
  echo -n -e "\\000"
end

function extraterm_postexec -e fish_postexec
  set -l status_backup $status
  echo -n -e -s "\\033" "&" $${this._extratermCookieName} ";3\\007"
  echo -n $status_backup
  echo -n -e "\\000"
end
`;
    }
    _formatCommand(commandName, commandSource) {
        return `function ${commandName}
  python3 -c '${this._escapeShellChars(commandSource)}' $argv
end
`;
    }
    _escapeShellChars(source) {
        return source.replace(/\\/g, '\\\\').replace(/'/g, "\\'");
    }
}
exports.FishScriptBuilder = FishScriptBuilder;
class BashScriptBuilder extends ScriptBuilder {
    build() {
        return [
            { type: 'text', text: CTRL_C },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty -echo; source /dev/stdin\n' },
            { type: 'wait', durationMilliseconds: 300 },
            ...super.build(),
            { type: 'text', text: 'echo "Shell integration is configured"\n' },
            { type: 'text', text: EOT },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty echo\n' },
        ];
    }
    _buildCookie() {
        return `export ${this._extratermCookieName}=${this._extratermCookieValue}`;
    }
    _buildShellReporting() {
        return `
postexec () {
  echo -n -e "\\033&$\{${this._extratermCookieName}\};3\\007"
  echo -n $1
  echo -n -e "\\000"
}
export PROMPT_COMMAND="postexec \\$?"

preexec () {
    echo -n -e "\\033&$\{${this._extratermCookieName}\};2;bash\\007"
    echo -n $1
    echo -n -e "\\000"
}

preexec_invoke_exec () {
    [ -n "$COMP_LINE" ] && return                     # do nothing if completing
    [ "$BASH_COMMAND" = "$PROMPT_COMMAND" ] && return # don't cause a preexec for $PROMPT_COMMAND
    local this_command=\`history 1\`; # obtain the command from the history
    preexec "$this_command"
}
trap 'preexec_invoke_exec' DEBUG
`;
    }
    _formatCommand(commandName, commandSource) {
        return `${commandName} () {
    python3 -c '${this._escapeShellChars(commandSource)}' "$@"
}
`;
    }
}
exports.BashScriptBuilder = BashScriptBuilder;
class ZshScriptBuilder extends ScriptBuilder {
    build() {
        return [
            { type: 'text', text: CTRL_C },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty -echo; source =(cat </dev/stdin)\n' },
            { type: 'wait', durationMilliseconds: 300 },
            ...super.build(),
            { type: 'text', text: 'echo "Shell integration is configured"\n' },
            { type: 'text', text: EOT },
            { type: 'wait', durationMilliseconds: 300 },
            { type: 'text', text: 'stty echo\n' },
        ];
    }
    _buildCookie() {
        return `export ${this._extratermCookieName}=${this._extratermCookieValue}`;
    }
    _buildShellReporting() {
        return `
if [[ ! "$PS1" =~ "${this._extratermCookieValue}" ]] ; then
    export PS1=\`echo -n -e "%{\\033&${this._extratermCookieValue};3\\007%}%?%{\\000%}\${PS1}"\`
fi

preexec () {
    echo -n -e "\\033&\${${this._extratermCookieName}};2;zsh\\007"
    echo -n $1
    echo -n -e "\\000"
}
`;
    }
    _formatCommand(commandName, commandSource) {
        return `${commandName} () {
    python3 -c '${this._escapeShellChars(commandSource)}' "$@"
}
`;
    }
    _escapeShellChars(source) {
        return source.replace(/'/g, "'\\''");
    }
}
exports.ZshScriptBuilder = ZshScriptBuilder;
class PythonFileFlattener {
    constructor(_baseDir) {
        this._baseDir = _baseDir;
    }
    readAndInlineCommand(commandName) {
        const commandSourceCode = fs.readFileSync(path.join(this._baseDir, commandName + ".py"), 'UTF-8');
        const lines = commandSourceCode.split('\n');
        const filteredLines = [];
        let inlineNextState = false;
        for (const line of lines) {
            if (!inlineNextState) {
                if (line.trim().startsWith('##@inline')) {
                    inlineNextState = true;
                }
                filteredLines.push(line);
            }
            else {
                filteredLines.push.apply(filteredLines, this._expandImport(line));
                inlineNextState = false;
            }
        }
        return this._removeDuplicateImports(filteredLines).join('\n');
    }
    _expandImport(line) {
        const importTarget = this._extractTargetFromImportLine(line);
        const librarySourceCode = fs.readFileSync(path.join(this._baseDir, importTarget + ".py"), 'UTF-8');
        const librarySourceLines = librarySourceCode.split('\n');
        return librarySourceLines;
    }
    _extractTargetFromImportLine(line) {
        if (line.startsWith('from ')) {
            const parts = line.slice('from '.length).split(' ');
            return parts[0];
        }
        return '';
    }
    _removeDuplicateImports(lines) {
        const result = [];
        const seenImports = new Set();
        for (const line of lines) {
            if (line.startsWith('import ')) {
                const importTarget = line.slice('import '.length);
                if (!seenImports.has(importTarget)) {
                    seenImports.add(importTarget);
                    result.push(line);
                }
            }
            else {
                result.push(line);
            }
        }
        return result;
    }
}
exports.PythonFileFlattener = PythonFileFlattener;
//# sourceMappingURL=ScriptBuilders.js.map