"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const path = require("path");
const _ = require("lodash");
const Logger_1 = require("./Logger");
const he = require("he");
const DomUtils = require("./DomUtils");
const CodeMirror = require("codemirror");
const ExtensionLoader_1 = require("./ExtensionLoader");
const Terminal_1 = require("./Terminal");
const ViewerElement_1 = require("./ViewerElement");
const TextViewer_1 = require("./viewers/TextViewer");
const OwnerTrackingList_1 = require("./utils/OwnerTrackingList");
const PopDownListPicker_1 = require("./gui/PopDownListPicker");
const PopDownNumberDialog_1 = require("./gui/PopDownNumberDialog");
const EmbeddedViewer_1 = require("./EmbeddedViewer");
const TerminalViewer_1 = require("./viewers/TerminalViewer");
class ExtensionManager {
    constructor() {
        this._log = null;
        this._extensionLoader = null;
        this._activeExtensions = [];
        this._extensionBridge = null;
        PopDownNumberDialog_1.PopDownNumberDialog.init();
        this._log = Logger_1.getLogger("ExtensionManager", this);
        this._extensionLoader = new ExtensionLoader_1.ExtensionLoader([path.join(__dirname, "../extensions")]);
        this._extensionBridge = new ExtensionBridge();
    }
    startUp() {
        this._extensionLoader.scan();
        for (const extensionInfo of this._extensionLoader.getExtensions()) {
            this._startExtension(extensionInfo);
        }
    }
    getExtensionBridge() {
        return this._extensionBridge;
    }
    _startExtension(extensionMetadata) {
        if (this._extensionLoader.load(extensionMetadata)) {
            try {
                const extensionContextImpl = this._extensionBridge.createExtensionSpecificContext(extensionMetadata);
                const extensionPublicApi = extensionMetadata.module.activate(extensionContextImpl);
                this._activeExtensions.push({ extensionMetadata, extensionPublicApi, extensionContextImpl });
            }
            catch (ex) {
                this._log.warn(`Exception occurred while starting extensions ${extensionMetadata.name}. ${ex}`);
            }
        }
    }
}
exports.ExtensionManager = ExtensionManager;
class ExtensionBridge {
    constructor() {
        this._log = null;
        this._numberInputDialog = null;
        this._listPicker = null;
        this.workspaceOnDidCreateTerminal = new OwnerTrackingEventListenerList();
        this.workspaceRegisterCommandsOnTerminal = new OwnerTrackingList_1.default();
        this.workspaceRegisterCommandsOnTextViewer = new OwnerTrackingList_1.default();
        this._log = Logger_1.getLogger("ExtensionBridge", this);
    }
    createExtensionSpecificContext(extensionMetadata) {
        return new ExtensionContextImpl(this, extensionMetadata);
    }
    workspaceGetTerminals() {
        return [];
    }
    getWorkspaceTerminalCommands(terminal) {
        return _.flatten(this.workspaceRegisterCommandsOnTerminal.mapWithOwner((ownerExtensionContext, registration) => {
            const terminalProxy = ownerExtensionContext.getTerminalProxy(terminal);
            const rawCommands = registration.commandLister(terminalProxy);
            const target = {
                executeCommand(commandId, options) {
                    const commandIdWithoutPrefix = commandId.slice(ownerExtensionContext.extensionMetadata.name.length + 1);
                    registration.commandExecutor(terminalProxy, commandIdWithoutPrefix, options);
                }
            };
            return this._formatCommands(rawCommands, target, ownerExtensionContext.extensionMetadata.name);
        }));
    }
    getWorkspaceTextViewerCommands(textViewer) {
        return _.flatten(this.workspaceRegisterCommandsOnTextViewer.mapWithOwner((ownerExtensionContext, registration) => {
            const textViewerProxy = ownerExtensionContext.getViewerProxy(textViewer);
            const rawCommands = registration.commandLister(textViewerProxy);
            const target = {
                executeCommand(commandId, options) {
                    const commandIdWithoutPrefix = commandId.slice(ownerExtensionContext.extensionMetadata.name.length + 1);
                    registration.commandExecutor(textViewerProxy, commandIdWithoutPrefix, options);
                }
            };
            return this._formatCommands(rawCommands, target, ownerExtensionContext.extensionMetadata.name);
        }));
    }
    _formatCommands(rawCommands, commandExecutor, commandPrefix) {
        const commands = [];
        for (const rawCommand of rawCommands) {
            commands.push({
                id: commandPrefix + '.' + rawCommand.id,
                group: rawCommand.group,
                iconLeft: rawCommand.iconLeft,
                iconRight: rawCommand.iconRight,
                label: rawCommand.label,
                shortcut: '',
                commandExecutor,
                commandArguments: rawCommand.commandArguments
            });
        }
        return commands;
    }
    showNumberInput(terminal, options) {
        let lastFocus = currentDeepFocusedViewerElement();
        if (lastFocus == null) {
            lastFocus = terminal;
        }
        if (this._numberInputDialog == null) {
            this._numberInputDialog = window.document.createElement(PopDownNumberDialog_1.PopDownNumberDialog.TAG_NAME);
            window.document.body.appendChild(this._numberInputDialog);
        }
        this._numberInputDialog.setTitlePrimary(options.title);
        this._numberInputDialog.setMinimum(options.minimum !== undefined ? options.minimum : Number.MIN_SAFE_INTEGER);
        this._numberInputDialog.setMaximum(options.maximum !== undefined ? options.maximum : Number.MAX_SAFE_INTEGER);
        this._numberInputDialog.setValue(options.value);
        const rect = terminal.getBoundingClientRect();
        this._numberInputDialog.open(rect.left, rect.top, rect.width, rect.height);
        this._numberInputDialog.focus();
        return new Promise((resolve, reject) => {
            const selectedHandler = (ev) => {
                this._numberInputDialog.removeEventListener('selected', selectedHandler);
                resolve(ev.detail.value == null ? undefined : ev.detail.value);
                lastFocus.focus();
            };
            this._numberInputDialog.addEventListener('selected', selectedHandler);
        });
    }
    showListPicker(terminal, options) {
        let lastFocus = currentDeepFocusedViewerElement();
        if (lastFocus == null) {
            lastFocus = terminal;
        }
        if (this._listPicker == null) {
            this._listPicker = window.document.createElement(PopDownListPicker_1.PopDownListPicker.TAG_NAME);
            this._listPicker.setFormatEntriesFunc((filteredEntries, selectedId, filterInputValue) => {
                return filteredEntries.map((entry) => {
                    return `<div class='CLASS_RESULT_ENTRY ${entry.id === selectedId ? PopDownListPicker_1.PopDownListPicker.CLASS_RESULT_SELECTED : ""}' ${PopDownListPicker_1.PopDownListPicker.ATTR_DATA_ID}='${entry.id}'>
            ${he.encode(entry.label)}
          </div>`;
                }).join("");
            });
            this._listPicker.setFilterAndRankEntriesFunc(this._listPickerFilterAndRankEntries.bind(this));
            window.document.body.appendChild(this._listPicker);
        }
        this._listPicker.setTitlePrimary(options.title);
        const convertedItems = options.items.map((item, index) => ({ id: "" + index, label: item }));
        this._listPicker.setEntries(convertedItems);
        this._listPicker.setSelected("" + options.selectedItemIndex);
        const rect = terminal.getBoundingClientRect();
        this._listPicker.open(rect.left, rect.top, rect.width, rect.height);
        this._listPicker.focus();
        return new Promise((resolve, reject) => {
            const selectedHandler = (ev) => {
                this._listPicker.removeEventListener('selected', selectedHandler);
                resolve(ev.detail.selected == null ? undefined : parseInt(ev.detail.selected, 10));
                lastFocus.focus();
            };
            this._listPicker.addEventListener('selected', selectedHandler);
        });
    }
    _listPickerFilterAndRankEntries(entries, filterText) {
        const lowerFilterText = filterText.toLowerCase().trim();
        const filtered = entries.filter((entry) => {
            return entry.label.toLowerCase().indexOf(lowerFilterText) !== -1;
        });
        const rankFunc = (entry, lowerFilterText) => {
            const lowerName = entry.label.toLowerCase();
            if (lowerName === lowerFilterText) {
                return 1000;
            }
            const pos = lowerName.indexOf(lowerFilterText);
            if (pos !== -1) {
                return 500 - pos; // Bias it for matches at the front of  the text.
            }
            return 0;
        };
        filtered.sort((a, b) => rankFunc(b, lowerFilterText) - rankFunc(a, lowerFilterText));
        return filtered;
    }
}
exports.ExtensionBridge = ExtensionBridge;
function currentDeepFocusedViewerElement() {
    const elements = DomUtils.activeNestedElements();
    const viewerElements = elements.filter(el => el instanceof ViewerElement_1.ViewerElement);
    return viewerElements.length === 0 ? null : viewerElements[0];
}
class OwnerTrackingEventListenerList extends OwnerTrackingList_1.default {
    emit(e) {
        this.forEach(thing => thing(e));
    }
}
class ExtensionContextImpl {
    constructor(extensionBridge, extensionMetadata) {
        this.extensionBridge = extensionBridge;
        this.extensionMetadata = extensionMetadata;
        this.workspace = null;
        this.codeMirrorModule = CodeMirror;
        this._tabProxyMap = new WeakMap();
        this._terminalProxyMap = new WeakMap();
        this._viewerProxyMap = new WeakMap();
        this.workspace = new WorkspaceProxy(this);
        this.logger = Logger_1.getLogger(extensionMetadata.name);
    }
    getTabProxy(terminal) {
        if (!this._tabProxyMap.has(terminal)) {
            this._tabProxyMap.set(terminal, new TerminalTabProxy(this, terminal));
        }
        return this._tabProxyMap.get(terminal);
    }
    getTerminalProxy(terminal) {
        if (!this._terminalProxyMap.has(terminal)) {
            this._terminalProxyMap.set(terminal, new TerminalProxy(this, terminal));
        }
        return this._terminalProxyMap.get(terminal);
    }
    getViewerProxy(viewer) {
        if (!this._viewerProxyMap.has(viewer)) {
            const proxy = this._createViewerProxy(viewer);
            if (proxy === null) {
                return null;
            }
            this._viewerProxyMap.set(viewer, proxy);
        }
        return this._viewerProxyMap.get(viewer);
    }
    _createViewerProxy(viewer) {
        if (viewer instanceof TerminalViewer_1.TerminalViewer) {
            return new TerminalOutputProxy(this, viewer);
        }
        if (viewer instanceof TextViewer_1.TextViewer) {
            return new TextViewerProxy(this, viewer);
        }
        if (viewer instanceof EmbeddedViewer_1.EmbeddedViewer) {
            return new FrameViewerProxy(this, viewer);
        }
        return null;
    }
}
class WorkspaceProxy {
    constructor(_extensionContextImpl) {
        this._extensionContextImpl = _extensionContextImpl;
    }
    getTerminals() {
        return this._extensionContextImpl.extensionBridge.workspaceGetTerminals()
            .map(terminal => this._extensionContextImpl.getTerminalProxy(terminal));
    }
    onDidCreateTerminal(listener) {
        return this._extensionContextImpl.extensionBridge.workspaceOnDidCreateTerminal.add(this._extensionContextImpl, listener);
    }
    registerCommandsOnTerminal(commandLister, commandExecutor) {
        return this._extensionContextImpl.extensionBridge.workspaceRegisterCommandsOnTerminal.add(this._extensionContextImpl, { commandLister, commandExecutor });
    }
    registerCommandsOnTextViewer(commandLister, commandExecutor) {
        return this._extensionContextImpl.extensionBridge.workspaceRegisterCommandsOnTextViewer.add(this._extensionContextImpl, { commandLister, commandExecutor });
    }
}
class TerminalTabProxy {
    constructor(_extensionContextImpl, _terminal) {
        this._extensionContextImpl = _extensionContextImpl;
        this._terminal = _terminal;
    }
    getTerminal() {
        return this._extensionContextImpl.getTerminalProxy(this._terminal);
    }
    showNumberInput(options) {
        return this._extensionContextImpl.extensionBridge.showNumberInput(this._terminal, options);
    }
    showListPicker(options) {
        return this._extensionContextImpl.extensionBridge.showListPicker(this._terminal, options);
    }
}
class TerminalProxy {
    constructor(_extensionContextImpl, _terminal) {
        this._extensionContextImpl = _extensionContextImpl;
        this._terminal = _terminal;
    }
    getTab() {
        return this._extensionContextImpl.getTabProxy(this._terminal);
    }
    type(text) {
        this._terminal.send(text);
    }
    getViewers() {
        return this._terminal.getViewerElements().map(viewer => this._extensionContextImpl.getViewerProxy(viewer));
    }
    getExtratermCookieValue() {
        return this._terminal.getExtratermCookieValue();
    }
    getExtratermCookieName() {
        return Terminal_1.EXTRATERM_COOKIE_ENV;
    }
}
class ViewerProxy {
    constructor(_extensionContextImpl, _viewer) {
        this._extensionContextImpl = _extensionContextImpl;
        this._viewer = _viewer;
    }
    getTab() {
        const terminal = this._getOwningEtTerminal();
        return terminal == null ? null : this._extensionContextImpl.getTabProxy(terminal);
    }
    _getOwningEtTerminal() {
        const path = DomUtils.nodePathToRoot(this._viewer);
        for (const node of path) {
            if (node instanceof Terminal_1.EtTerminal) {
                return node;
            }
        }
        return null;
    }
    getOwningTerminal() {
        const terminal = this._getOwningEtTerminal();
        return terminal == null ? null : this._extensionContextImpl.getTerminalProxy(terminal);
    }
}
class TerminalOutputProxy extends ViewerProxy {
    constructor(_extensionContextImpl, _terminalViewer) {
        super(_extensionContextImpl, _terminalViewer);
        this._extensionContextImpl = _extensionContextImpl;
        this._terminalViewer = _terminalViewer;
        this.viewerType = 'terminal-output';
    }
    isLive() {
        return this._terminalViewer.getEmulator() != null;
    }
}
class FrameViewerProxy extends ViewerProxy {
    constructor(_extensionContextImpl, _embeddedViewer) {
        super(_extensionContextImpl, _embeddedViewer);
        this._extensionContextImpl = _extensionContextImpl;
        this._embeddedViewer = _embeddedViewer;
        this.viewerType = 'frame';
    }
    getContents() {
        const viewerElement = this._embeddedViewer.getViewerElement();
        if (viewerElement !== null) {
            return this._extensionContextImpl.getViewerProxy(viewerElement);
        }
        return null;
    }
}
class TextViewerProxy extends ViewerProxy {
    constructor(_extensionContextImpl, _textViewer) {
        super(_extensionContextImpl, _textViewer);
        this._textViewer = _textViewer;
        this.viewerType = 'text';
    }
    getTabSize() {
        return this._textViewer.getTabSize();
    }
    setTabSize(size) {
        this._textViewer.setTabSize(size);
    }
    getMimeType() {
        return this._textViewer.getMimeType();
    }
    setMimeType(mimeType) {
        this._textViewer.setMimeType(mimeType);
    }
}
//# sourceMappingURL=ExtensionManager.js.map