"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Electron = require("electron");
const ElectronMenu = Electron.remote.Menu;
const ElectronMenuItem = Electron.remote.MenuItem;
const SourceMapSupport = require("source-map-support");
const _ = require("lodash");
const Logger_1 = require("./Logger");
const Messages = require("./WindowMessages");
const WebIpc = require("./WebIpc");
const ContextMenu_1 = require("./gui/ContextMenu");
const MenuItem_1 = require("./gui/MenuItem");
const DropDown_1 = require("./gui/DropDown");
const CheckboxMenuItem_1 = require("./gui/CheckboxMenuItem");
const PopDownListPicker_1 = require("./gui/PopDownListPicker");
const ResizeRefreshElementBase = require("./ResizeRefreshElementBase");
const CommandPaletteRequestTypes_1 = require("./CommandPaletteRequestTypes");
const CommandPaletteFunctions_1 = require("./CommandPaletteFunctions");
const GeneralEvents_1 = require("./GeneralEvents");
const ExtensionManager_1 = require("./ExtensionManager");
const MainWebUi_1 = require("./MainWebUi");
const Terminal_1 = require("./Terminal");
const DomUtils = require("./DomUtils");
const Util = require("./gui/Util");
const TextViewer_1 = require("./viewers/TextViewer");
const ResizeCanary_1 = require("./ResizeCanary");
const config = require("./Config");
const ThemeTypes = require("./Theme");
const ThemeConsumer = require("./ThemeConsumer");
const keybindingmanager = require("./KeyBindingManager");
SourceMapSupport.install();
const PLUGINS_DIRECTORY = "plugins";
const PALETTE_GROUP = "mainweb";
const MENU_ITEM_SETTINGS = 'settings';
const MENU_ITEM_KEY_BINDINGS = 'key_bindings';
const MENU_ITEM_DEVELOPER_TOOLS = 'developer_tools';
const MENU_ITEM_ABOUT = 'about';
const MENU_ITEM_RELOAD_CSS = 'reload_css';
const ID_COMMAND_PALETTE = "ID_COMMAND_PALETTE";
const ID_MAIN_MENU = "ID_MAIN_MENU";
const ID_MENU_BUTTON = "ID_MENU_BUTTON";
const CLASS_MAIN_DRAGGING = "CLASS_MAIN_DRAGGING";
const CLASS_MAIN_NOT_DRAGGING = "CLASS_MAIN_NOT_DRAGGING";
const _log = Logger_1.getLogger("mainweb");
/**
 * This module has control of the window and is responsible for
 * starting up the main component and handling the window directly.
 */
let terminalIdCounter = 0;
let keyBindingManager = null;
let themes;
let mainWebUi = null;
let configManager = null;
let extensionManager = null;
/**
 *
 */
function startUp() {
    if (process.platform === "darwin") {
        setupOSXEmptyMenus();
    }
    startUpTheming();
    startUpWebIpc();
    // Get the Config working.
    configManager = new ConfigManagerImpl();
    keyBindingManager = new KeyBindingManagerImpl(); // depends on the config.
    const themePromise = WebIpc.requestConfig().then((msg) => {
        return handleConfigMessage(msg);
    });
    // Get the config and theme info in and then continue starting up.
    const allPromise = Promise.all([themePromise, WebIpc.requestThemeList().then(handleThemeListMessage)]);
    allPromise.then(loadFontFaces)
        .then(() => {
        startUpComponents();
        const doc = window.document;
        doc.body.classList.remove("preparing");
        doc.body.innerHTML = ""; // Remove the old contents.
        doc.body.classList.add(CLASS_MAIN_NOT_DRAGGING);
        startUpMainWebUi();
        startUpMainMenu();
        startUpResizeCanary();
        startUpCommandPalette();
        startUpWindowEvents();
        if (process.platform === "darwin") {
            setupOSXMenus(mainWebUi);
        }
        startUpExtensions();
        mainWebUi.newTerminalTab();
        mainWebUi.focus();
        window.focus();
    });
}
exports.startUp = startUp;
function startUpTheming() {
    // Theme control for the window level.
    const topThemeable = {
        setThemeCssMap(cssMap) {
            document.getElementById('THEME_STYLE').textContent =
                cssMap.get(ThemeTypes.CssFile.GUI_CONTROLS) + "\n" +
                    cssMap.get(ThemeTypes.CssFile.FONT_AWESOME) + "\n" +
                    cssMap.get(ThemeTypes.CssFile.TOP_WINDOW) + "\n" +
                    cssMap.get(ThemeTypes.CssFile.TERMINAL_VARS) + "\n" +
                    cssMap.get(ThemeTypes.CssFile.THEME_VARS);
        }
    };
    ThemeConsumer.registerThemeable(topThemeable);
}
function startUpWebIpc() {
    WebIpc.start();
    // Default handling for config messages.
    WebIpc.registerDefaultHandler(1 /* CONFIG */, handleConfigMessage);
    // Default handling for theme messages.
    WebIpc.registerDefaultHandler(5 /* THEME_LIST */, handleThemeListMessage);
    WebIpc.registerDefaultHandler(7 /* THEME_CONTENTS */, handleThemeContentsMessage);
    WebIpc.registerDefaultHandler(16 /* DEV_TOOLS_STATUS */, handleDevToolsStatus);
    WebIpc.registerDefaultHandler(19 /* CLIPBOARD_READ */, handleClipboardRead);
}
function loadFontFaces() {
    // Next phase is wait for the fonts to load.
    const fontPromises = [];
    window.document.fonts.forEach((font) => {
        if (font.status !== 'loaded' && font.status !== 'loading') {
            fontPromises.push(font.load());
        }
    });
    return Promise.all(fontPromises);
}
function startUpComponents() {
    // Fonts are loaded, continue.
    ContextMenu_1.ContextMenu.init();
    MenuItem_1.MenuItem.init();
    DropDown_1.DropDown.init();
    MainWebUi_1.MainWebUi.init();
    CheckboxMenuItem_1.CheckboxMenuItem.init();
    PopDownListPicker_1.PopDownListPicker.init();
    ResizeCanary_1.ResizeCanary.init();
}
function startUpMainWebUi() {
    mainWebUi = window.document.createElement(MainWebUi_1.MainWebUi.TAG_NAME);
    config.injectConfigManager(mainWebUi, configManager);
    keybindingmanager.injectKeyBindingManager(mainWebUi, keyBindingManager);
    mainWebUi.innerHTML = `<div class="tab_bar_rest">
    <div class="space"></div>
    <button id="${ID_MENU_BUTTON}" class="btn btn-quiet"><i class="fa fa-bars"></i></button>
    </div>`;
    mainWebUi.setThemes(themes);
    window.document.body.appendChild(mainWebUi);
    // Detect when the last tab has closed.
    mainWebUi.addEventListener(MainWebUi_1.MainWebUi.EVENT_TAB_CLOSED, (ev) => {
        if (mainWebUi.getTabCount() === 0) {
            WebIpc.windowCloseRequest();
        }
    });
    // Update the window title on request.
    mainWebUi.addEventListener(MainWebUi_1.MainWebUi.EVENT_TITLE, (ev) => {
        window.document.title = "Extraterm - " + ev.detail.title;
    });
    mainWebUi.addEventListener(MainWebUi_1.MainWebUi.EVENT_MINIMIZE_WINDOW_REQUEST, () => {
        WebIpc.windowMinimizeRequest();
    });
    mainWebUi.addEventListener(MainWebUi_1.MainWebUi.EVENT_MAXIMIZE_WINDOW_REQUEST, () => {
        WebIpc.windowMaximizeRequest();
    });
    mainWebUi.addEventListener(MainWebUi_1.MainWebUi.EVENT_CLOSE_WINDOW_REQUEST, () => {
        WebIpc.windowCloseRequest();
    });
    mainWebUi.addEventListener(GeneralEvents_1.EVENT_DRAG_STARTED, (ev) => {
        window.document.body.classList.add(CLASS_MAIN_DRAGGING);
        window.document.body.classList.remove(CLASS_MAIN_NOT_DRAGGING);
    });
    mainWebUi.addEventListener(GeneralEvents_1.EVENT_DRAG_ENDED, (ev) => {
        window.document.body.classList.remove(CLASS_MAIN_DRAGGING);
        window.document.body.classList.add(CLASS_MAIN_NOT_DRAGGING);
    });
    mainWebUi.addEventListener(CommandPaletteRequestTypes_1.EVENT_COMMAND_PALETTE_REQUEST, handleCommandPaletteRequest);
}
function startUpMainMenu() {
    const contextMenuFragment = DomUtils.htmlToFragment(`
    <${ContextMenu_1.ContextMenu.TAG_NAME} id="${ID_MAIN_MENU}">
        <${MenuItem_1.MenuItem.TAG_NAME} icon="wrench" name="${MENU_ITEM_SETTINGS}">Settings</${MenuItem_1.MenuItem.TAG_NAME}>
        <${MenuItem_1.MenuItem.TAG_NAME} icon="keyboard-o" name="${MENU_ITEM_KEY_BINDINGS}">Key Bindings</${MenuItem_1.MenuItem.TAG_NAME}>
        <${CheckboxMenuItem_1.CheckboxMenuItem.TAG_NAME} icon="cogs" id="${MENU_ITEM_DEVELOPER_TOOLS}" name="developer_tools">Developer Tools</${CheckboxMenuItem_1.CheckboxMenuItem.TAG_NAME}>
        <${MenuItem_1.MenuItem.TAG_NAME} icon="lightbulb-o" name="${MENU_ITEM_ABOUT}">About</${MenuItem_1.MenuItem.TAG_NAME}>
    </${ContextMenu_1.ContextMenu.TAG_NAME}>
  `);
    window.document.body.appendChild(contextMenuFragment);
    const mainMenu = window.document.getElementById(ID_MAIN_MENU);
    mainMenu.addEventListener('selected', (ev) => {
        executeMenuCommand(ev.detail.name);
    });
    const menuButton = document.getElementById(ID_MENU_BUTTON);
    menuButton.addEventListener('click', () => {
        const contextMenu = document.getElementById(ID_MAIN_MENU);
        contextMenu.openAround(menuButton);
    });
}
function startUpResizeCanary() {
    // A special element for tracking when terminal fonts are effectively changed in the DOM.
    const resizeCanary = window.document.createElement(ResizeCanary_1.ResizeCanary.TAG_NAME);
    resizeCanary.setCss(`
  font-family: var(--terminal-font);
  font-size: var(--default-terminal-font-size);
`);
    window.document.body.appendChild(resizeCanary);
    resizeCanary.addEventListener('resize', () => {
        mainWebUi.refresh(ResizeRefreshElementBase.RefreshLevel.COMPLETE);
    });
}
function startUpWindowEvents() {
    // Make sure something sensible is focussed if the window gets the focus.
    window.addEventListener('focus', () => {
        mainWebUi.focus();
    });
    window.addEventListener('resize', () => {
        if (mainWebUi !== null) {
            mainWebUi.refresh(ResizeRefreshElementBase.RefreshLevel.RESIZE);
        }
    });
    window.document.addEventListener('mousedown', (ev) => {
        if (ev.which === 2) {
            WebIpc.clipboardReadRequest();
            // This is needed to stop the autoscroll blob from appearing on Windows.
            ev.preventDefault();
            ev.stopPropagation();
        }
    });
    window.document.addEventListener('dragover', (ev) => {
        ev.preventDefault();
        ev.stopPropagation();
    });
    window.document.addEventListener('drop', (ev) => {
        ev.preventDefault();
        ev.stopPropagation();
    });
}
function startUpExtensions() {
    extensionManager = new ExtensionManager_1.ExtensionManager();
    extensionManager.startUp();
}
function executeMenuCommand(command) {
    if (command === MENU_ITEM_DEVELOPER_TOOLS) {
        // Unflip what the user did to the state of the developer tools check box for a moment.
        // Let executeCommand() toggle the checkbox itself. 
        const developerToolMenu = document.getElementById("developer_tools");
        const devToolsOpen = Util.toBoolean(developerToolMenu.getAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED));
        developerToolMenu.setAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED, "" + (!devToolsOpen));
    }
    return executeCommand(command);
}
function executeCommand(commandId, options) {
    switch (commandId) {
        case MENU_ITEM_SETTINGS:
            mainWebUi.openSettingsTab();
            break;
        case MENU_ITEM_KEY_BINDINGS:
            mainWebUi.openKeyBindingsTab();
            break;
        case MENU_ITEM_DEVELOPER_TOOLS:
            const developerToolMenu = document.getElementById("developer_tools");
            const devToolsOpen = Util.toBoolean(developerToolMenu.getAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED));
            developerToolMenu.setAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED, "" + (!devToolsOpen));
            WebIpc.devToolsRequest(!devToolsOpen);
            break;
        case MENU_ITEM_ABOUT:
            mainWebUi.openAboutTab();
            break;
        case MENU_ITEM_RELOAD_CSS:
            reloadThemeContents();
            break;
        default:
            return false;
    }
    return true;
}
function setupOSXEmptyMenus() {
    const template = [{
            label: "Extraterm",
        }];
    const emptyTopMenu = ElectronMenu.buildFromTemplate(template);
    ElectronMenu.setApplicationMenu(emptyTopMenu);
}
function setupOSXMenus(mainWebUi) {
    const template = [{
            label: "Extraterm",
            submenu: [
                {
                    label: 'About Extraterm',
                    click(item, focusedWindow) {
                        mainWebUi.openAboutTab();
                    },
                },
                {
                    type: 'separator'
                },
                {
                    label: 'Preferences...',
                    click(item, focusedWindow) {
                        mainWebUi.openSettingsTab();
                    },
                },
                {
                    label: 'Key Bindings...',
                    click(item, focusedWindow) {
                        mainWebUi.openKeyBindingsTab();
                    },
                },
                {
                    type: 'separator'
                },
                {
                    label: 'Quit',
                    click(item, focusedWindow) {
                        WebIpc.windowCloseRequest();
                    },
                    accelerator: 'Command+Q'
                }
            ]
        }];
    const topMenu = ElectronMenu.buildFromTemplate(template);
    ElectronMenu.setApplicationMenu(topMenu);
}
function handleConfigMessage(msg) {
    const configMessage = msg;
    const oldConfiguration = configManager.getConfig();
    const config = configMessage.config;
    configManager.setNewConfig(config);
    return setupConfiguration(oldConfiguration, config);
}
function handleThemeListMessage(msg) {
    const themesMessage = msg;
    themes = themesMessage.themeInfo;
}
function handleThemeContentsMessage(msg) {
    const themeContentsMessage = msg;
    if (themeContentsMessage.success) {
        const cssFileMap = new Map();
        themeContentsMessage.cssFileList.forEach((cssFile) => {
            cssFileMap.set(cssFile, themeContentsMessage.themeContents.cssFiles[ThemeTypes.cssFileNameBase(cssFile)]);
        });
        // Distribute the CSS files to the classes which want them.
        ThemeConsumer.updateCss(cssFileMap);
    }
    else {
        themeContentsError(themeContentsMessage);
    }
}
function themeContentsError(themeContentsMessage) {
    // Something went wrong.
    _log.warn(themeContentsMessage.errorMessage);
    if (themeContentsMessage.themeIdList.every(id => id === ThemeTypes.FALLBACK_UI_THEME)) {
        // Error occurred while trying to generate the default themes.
        window.alert("Something has gone wrong. The default theme couldn't be generated. Sorry.");
    }
    else {
        _log.warn("Attempting to use the default theme.");
        window.alert("Something has gone wrong while generating the theme. The default theme will be tried.");
        requestThemeContents(ThemeTypes.FALLBACK_TERMINAL_THEME, ThemeTypes.FALLBACK_SYNTAX_THEME, ThemeTypes.FALLBACK_UI_THEME);
    }
}
function handleDevToolsStatus(msg) {
    const devToolsStatusMessage = msg;
    const developerToolMenu = document.getElementById("developer_tools");
    if (developerToolMenu === null) {
        return;
    }
    developerToolMenu.setAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED, "" + devToolsStatusMessage.open);
}
function handleClipboardRead(msg) {
    const clipboardReadMessage = msg;
    mainWebUi.pasteText(clipboardReadMessage.text);
}
//-------------------------------------------------------------------------
/**
 *
 */
function setupConfiguration(oldConfig, newConfig) {
    const keyBindingContexts = keybindingmanager.loadKeyBindingsFromObject(newConfig.systemConfig.keyBindingsContexts, process.platform);
    if (!keyBindingContexts.equals(keyBindingManager.getKeyBindingContexts())) {
        keyBindingManager.setKeyBindingContexts(keyBindingContexts);
    }
    if (oldConfig === null || oldConfig.terminalFontSize !== newConfig.terminalFontSize ||
        oldConfig.terminalFont !== newConfig.terminalFont) {
        const matchingFonts = newConfig.systemConfig.availableFonts.filter((font) => font.postscriptName === newConfig.terminalFont);
        setCssVars(newConfig.terminalFont, matchingFonts[0].path, newConfig.terminalFontSize);
    }
    if (oldConfig === null || oldConfig.themeTerminal !== newConfig.themeTerminal ||
        oldConfig.themeSyntax !== newConfig.themeSyntax ||
        oldConfig.themeGUI !== newConfig.themeGUI) {
        return requestThemeContents(newConfig.themeTerminal, newConfig.themeSyntax, newConfig.themeGUI);
    }
    // no-op promise.
    return new Promise((resolve, cancel) => { resolve(); });
}
function requestThemeContents(themeTerminal, themeSyntax, themeGUI) {
    const terminalThemeIdList = [themeTerminal, ThemeTypes.FALLBACK_TERMINAL_THEME];
    const syntaxThemeIdList = [themeSyntax, ThemeTypes.FALLBACK_SYNTAX_THEME];
    const uiThemeIdList = [themeGUI, ThemeTypes.FALLBACK_UI_THEME];
    const cssFileMap = new Map();
    return WebIpc.requestThemeContents(terminalThemeIdList, ThemeTypes.TerminalCssFiles)
        .then((result) => {
        if (result.success) {
            ThemeTypes.TerminalCssFiles.forEach((cssFile) => {
                const key = ThemeTypes.cssFileNameBase(cssFile);
                cssFileMap.set(cssFile, result.themeContents.cssFiles[key]);
            });
        }
        return WebIpc.requestThemeContents(syntaxThemeIdList, ThemeTypes.SyntaxCssFiles);
    }, themeContentsError)
        .then((result) => {
        if (result.success) {
            ThemeTypes.SyntaxCssFiles.forEach((cssFile) => {
                const key = ThemeTypes.cssFileNameBase(cssFile);
                cssFileMap.set(cssFile, result.themeContents.cssFiles[key]);
            });
        }
        return WebIpc.requestThemeContents(uiThemeIdList, ThemeTypes.UiCssFiles);
    }, themeContentsError)
        .then((result) => {
        if (result.success) {
            ThemeTypes.UiCssFiles.forEach((cssFile) => {
                const key = ThemeTypes.cssFileNameBase(cssFile);
                cssFileMap.set(cssFile, result.themeContents.cssFiles[key]);
            });
        }
        // Distribute the CSS files to the classes which want them.
        ThemeConsumer.updateCss(cssFileMap);
    }, themeContentsError);
}
function reloadThemeContents() {
    const config = configManager.getConfig();
    requestThemeContents(config.themeTerminal, config.themeSyntax, config.themeGUI);
}
function setCssVars(fontName, fontPath, terminalFontSize) {
    const fontCssName = fontName.replace(/\W/g, "_");
    document.getElementById('CSS_VARS').textContent =
        `
    @font-face {
      font-family: "${fontCssName}";
      src: url("${fontPath}");
    }

    :root {
      --default-terminal-font-size: ${terminalFontSize}px;
      --terminal-font: "${fontCssName}";
    }
    `;
}
//-----------------------------------------------------------------------
//
//   #####                                               ######                                          
//  #     #  ####  #    # #    #   ##   #    # #####     #     #   ##   #      ###### ##### ##### ###### 
//  #       #    # ##  ## ##  ##  #  #  ##   # #    #    #     #  #  #  #      #        #     #   #      
//  #       #    # # ## # # ## # #    # # #  # #    #    ######  #    # #      #####    #     #   #####  
//  #       #    # #    # #    # ###### #  # # #    #    #       ###### #      #        #     #   #      
//  #     # #    # #    # #    # #    # #   ## #    #    #       #    # #      #        #     #   #      
//   #####   ####  #    # #    # #    # #    # #####     #       #    # ###### ######   #     #   ###### 
//                                                                                                      
//-----------------------------------------------------------------------
let commandPaletteRequestSource = null;
let commandPaletteRequestEntries = null;
function startUpCommandPalette() {
    const doc = window.document;
    // Command palette
    const commandPalette = doc.createElement(PopDownListPicker_1.PopDownListPicker.TAG_NAME);
    commandPalette.id = ID_COMMAND_PALETTE;
    commandPalette.setTitlePrimary("Command Palette");
    commandPalette.setTitleSecondary("Ctrl+Shift+P");
    commandPalette.setFilterAndRankEntriesFunc(CommandPaletteFunctions_1.commandPaletteFilterEntries);
    commandPalette.setFormatEntriesFunc(CommandPaletteFunctions_1.commandPaletteFormatEntries);
    commandPalette.addExtraCss([ThemeTypes.CssFile.GUI_COMMANDPALETTE]);
    doc.body.appendChild(commandPalette);
    commandPalette.addEventListener('selected', handleCommandPaletteSelected);
}
function handleCommandPaletteRequest(ev) {
    const path = ev.composedPath();
    const requestCommandableStack = path.filter(el => CommandPaletteRequestTypes_1.isCommandable(el));
    DomUtils.doLater(() => {
        const commandableStack = [...requestCommandableStack, { executeCommand, getCommandPaletteEntries }];
        const firstCommandable = commandableStack[0];
        if (firstCommandable instanceof HTMLElement) {
            commandPaletteRequestSource = firstCommandable;
        }
        commandPaletteRequestEntries = _.flatten(commandableStack.map(commandable => {
            let result = commandable.getCommandPaletteEntries(commandableStack);
            if (commandable instanceof Terminal_1.EtTerminal) {
                result = [...result, ...extensionManager.getExtensionBridge().getWorkspaceTerminalCommands(commandable)];
            }
            else if (commandable instanceof TextViewer_1.TextViewer) {
                result = [...result, ...extensionManager.getExtensionBridge().getWorkspaceTextViewerCommands(commandable)];
            }
            return result;
        }));
        const paletteEntries = commandPaletteRequestEntries.map((entry, index) => {
            return {
                id: "" + index,
                group: entry.group,
                iconLeft: entry.iconLeft,
                iconRight: entry.iconRight,
                label: entry.label,
                shortcut: entry.shortcut
            };
        });
        const commandPalette = document.getElementById(ID_COMMAND_PALETTE);
        const shortcut = keyBindingManager.getKeyBindingContexts().context("main-ui").mapCommandToKeyBinding("openCommandPalette");
        commandPalette.setTitleSecondary(shortcut !== null ? shortcut : "");
        commandPalette.setEntries(paletteEntries);
        const contextElement = requestCommandableStack[requestCommandableStack.length - 2];
        if (contextElement instanceof HTMLElement) {
            let rect = { left: 0, top: 0, width: 500, height: 500, right: 500, bottom: 500 };
            if (contextElement != null) {
                rect = contextElement.getBoundingClientRect();
            }
            commandPalette.open(rect.left, rect.top, rect.width, rect.height);
            commandPalette.focus();
        }
    });
}
function getCommandPaletteEntries(commandableStack) {
    const developerToolMenu = document.getElementById("developer_tools");
    const devToolsOpen = Util.toBoolean(developerToolMenu.getAttribute(CheckboxMenuItem_1.CheckboxMenuItem.ATTR_CHECKED));
    const commandExecutor = { executeCommand };
    const commandList = [
        { id: MENU_ITEM_SETTINGS, group: PALETTE_GROUP, iconRight: "wrench", label: "Settings", commandExecutor },
        { id: MENU_ITEM_KEY_BINDINGS, group: PALETTE_GROUP, iconRight: "keyboard-o", label: "Key Bindings", commandExecutor },
        { id: MENU_ITEM_DEVELOPER_TOOLS, group: PALETTE_GROUP, iconLeft: devToolsOpen ? "check-square-o" : "square-o", iconRight: "cogs", label: "Developer Tools", commandExecutor },
        { id: MENU_ITEM_RELOAD_CSS, group: PALETTE_GROUP, iconRight: "refresh", label: "Reload Theme", commandExecutor },
        { id: MENU_ITEM_ABOUT, group: PALETTE_GROUP, iconRight: "lightbulb-o", label: "About", commandExecutor },
    ];
    return commandList;
}
function handleCommandPaletteSelected(ev) {
    const commandPalette = document.getElementById(ID_COMMAND_PALETTE);
    commandPalette.close();
    if (commandPaletteRequestSource !== null) {
        commandPaletteRequestSource.focus();
    }
    const selectedId = ev.detail.selected;
    if (selectedId !== null) {
        const commandIndex = Number.parseInt(selectedId);
        const commandEntry = commandPaletteRequestEntries[commandIndex];
        DomUtils.doLater(() => {
            commandEntry.commandExecutor.executeCommand(commandEntry.id, commandEntry.commandArguments);
            commandPaletteRequestSource = null;
            commandPaletteRequestEntries = null;
        });
    }
}
class ConfigManagerImpl {
    constructor() {
        this._config = null;
        this._listenerList = []; // Immutable list
    }
    registerChangeListener(key, onChange) {
        this._listenerList = [...this._listenerList, { key, onChange }];
    }
    unregisterChangeListener(key) {
        this._listenerList = this._listenerList.filter((tup) => tup.key !== key);
    }
    getConfig() {
        return this._config;
    }
    setConfig(newConfig) {
        WebIpc.sendConfig(newConfig);
    }
    /**
     * Set a new configuration object as the application wide
     */
    setNewConfig(newConfig) {
        this._config = newConfig;
        const listenerList = this._listenerList;
        for (const tup of listenerList) {
            tup.onChange();
        }
    }
}
class KeyBindingManagerImpl {
    constructor() {
        this._keyBindingContexts = null;
        this._listenerList = []; // Immutable list
    }
    getKeyBindingContexts() {
        return this._keyBindingContexts;
    }
    setKeyBindingContexts(newKeyBindingContexts) {
        this._keyBindingContexts = newKeyBindingContexts;
        const listenerList = this._listenerList;
        for (const tup of listenerList) {
            tup.onChange();
        }
    }
    /**
     * Register a listener to hear when the key bindings change.
     *
     * @param key an opaque object which is used to identify this registration.
     * @param onChange the function to call when the config changes.
     */
    registerChangeListener(key, onChange) {
        this._listenerList = [...this._listenerList, { key, onChange }];
    }
    /**
     * Unregister a listener.
     *
     * @param key the same opaque object which was used during registerChangeListener().
     */
    unregisterChangeListener(key) {
        this._listenerList = this._listenerList.filter((tup) => tup.key !== key);
    }
}
//# sourceMappingURL=MainWeb.js.map