"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const child_process = require("child_process");
const _ = require("lodash");
const fs = require("fs");
const path = require("path");
const Logger_1 = require("./Logger");
const SourceDir = require("./SourceDir");
const DEBUG_FINE = false;
const _log = Logger_1.getLogger("ptyproxy");
const TYPE_CREATE = "create";
const TYPE_CREATED = "created";
const TYPE_WRITE = "write";
const TYPE_OUTPUT = "output";
const TYPE_RESIZE = "resize";
const TYPE_CLOSED = "closed";
const TYPE_TERMINATE = "terminate";
const TYPE_PERMIT_DATA_SIZE = "permit-data-size";
const NULL_ID = -1;
class ProxyPty {
    constructor(writeFunc) {
        this._id = NULL_ID;
        this._dataListener = null;
        this._exitListener = null;
        this._writeFunc = null;
        // Pre-open write queue.
        this._writeQueue = [];
        this._live = true;
        this._writeFunc = writeFunc;
    }
    getId() {
        return this._id;
    }
    setId(id) {
        this._id = id;
        if (this._live) {
            this._writeQueue.forEach((msg) => {
                msg.id = this._id;
                this._writeFunc(this._id, msg);
            });
            this._writeQueue = [];
        }
    }
    _writeMessage(id, msg) {
        if (this._live) {
            if (this._id === -1) {
                // We don't know what the ID of the pty in the proxy is.
                // Queue up this message for later.
                this._writeQueue.push(msg);
            }
            else {
                this._writeFunc(this._id, msg);
            }
        }
    }
    write(data) {
        const msg = { type: TYPE_WRITE, id: this._id, data: data };
        this._writeMessage(this._id, msg);
    }
    resize(cols, rows) {
        const msg = { type: TYPE_RESIZE, id: this._id, rows: rows, columns: cols };
        this._writeMessage(this._id, msg);
    }
    onData(callback) {
        this._dataListener = callback;
    }
    permittedDataSize(size) {
        const msg = { type: TYPE_PERMIT_DATA_SIZE, id: this._id, size: size };
        this._writeMessage(this._id, msg);
    }
    data(data) {
        if (this._live && this._dataListener !== null) {
            this._dataListener(data);
        }
    }
    onExit(callback) {
        this._exitListener = callback;
    }
    exit() {
        if (this._exitListener !== null) {
            this._live = false;
            this._exitListener();
        }
    }
    destroy() {
    }
}
function findCygwinPython(cygwinDir) {
    const binDir = path.join(cygwinDir, 'bin');
    _log.info("Cygwin bin directory is ", binDir);
    if (fs.existsSync(binDir)) {
        const pythonRegexp = /^python3.*m\.exe$/;
        const binContents = fs.readdirSync(binDir);
        const pythons = binContents.filter(name => pythonRegexp.test(name));
        return pythons.length !== 0 ? path.join(binDir, pythons[0]) : null;
    }
    return null;
}
function factory(config) {
    const ptys = [];
    const sessionProfile = config.expandedProfiles[0];
    const pythonExe = findCygwinPython(sessionProfile.cygwinDir);
    if (pythonExe !== null) {
        _log.info("Found Python 3 exe at ", pythonExe);
    }
    else {
        const msg = "Unable to find Python 3 exe.";
        _log.severe("Unable to find Python 3 exe.");
        throw new Error(msg);
    }
    const serverEnv = _.clone(process.env);
    serverEnv["PYTHONIOENCODING"] = "utf-8:ignore";
    const proxy = child_process.spawn(pythonExe, [path.join(SourceDir.path, 'python/ptyserver2.py')], { env: serverEnv });
    let messageBuffer = "";
    proxy.stdout.on('data', function (data) {
        if (DEBUG_FINE) {
            _log.debug("main <<< server : ", data);
        }
        messageBuffer = messageBuffer + data.toString('utf8');
        processMessageBuffer();
    });
    proxy.stderr.on('data', function (data) {
        _log.warn('ptyproxy process stderr: ', data.toString());
    });
    proxy.on('close', function (code) {
        if (DEBUG_FINE) {
            _log.debug('bridge process closed with code: ', code);
        }
    });
    proxy.on('exit', function (code) {
        if (DEBUG_FINE) {
            _log.debug('bridge process exited with code: ', code);
        }
    });
    proxy.on('error', (err) => {
        _log.severe("Failed to start process " + pythonExe + ". ", err);
    });
    function processMessageBuffer() {
        while (true) {
            const end = messageBuffer.indexOf('\n');
            if (end !== -1) {
                const msgString = messageBuffer.slice(0, end);
                messageBuffer = messageBuffer.slice(end + 1);
                const msg = JSON.parse(msgString);
                processMessage(msg);
            }
            else {
                break;
            }
        }
    }
    function processMessage(msg) {
        const msgType = msg.type;
        if (msgType === TYPE_CREATED) {
            const createdPtyMsg = msg;
            for (let i = 0; i < ptys.length; i++) {
                const pty = ptys[i];
                if (pty.getId() === NULL_ID) {
                    pty.setId(createdPtyMsg.id);
                    break;
                }
            }
            return;
        }
        if (msgType === TYPE_OUTPUT) {
            const outputMsg = msg;
            const pty = findPtyById(outputMsg.id);
            if (pty !== null) {
                pty.data(outputMsg.data);
            }
        }
        if (msgType === TYPE_CLOSED) {
            const closedMsg = msg;
            const pty = findPtyById(closedMsg.id);
            if (pty !== null) {
                pty.exit();
            }
        }
    }
    function findPtyById(id) {
        for (let i = 0; i < ptys.length; i++) {
            if (ptys[i].getId() === id) {
                return ptys[i];
            }
        }
        return null;
    }
    function sendMessage(id, msg) {
        const msgText = JSON.stringify(msg);
        proxy.stdin.write(msgText + "\n", 'utf8');
    }
    function spawn(file, args, opt) {
        let rows = 24;
        let columns = 80;
        if (DEBUG_FINE) {
            _log.debug("ptyproxy spawn file: ", file);
        }
        if (opt !== undefined) {
            rows = opt.rows !== undefined ? opt.rows : rows;
            columns = opt.cols !== undefined ? opt.cols : columns;
        }
        const pty = new ProxyPty(sendMessage);
        ptys.push(pty);
        const msg = { type: TYPE_CREATE, argv: [file, ...args], rows: rows, columns: columns, id: -1, env: opt.env };
        sendMessage(null, msg);
        return pty;
    }
    function destroy() {
        const msg = { type: TYPE_TERMINATE, id: -1 };
        sendMessage(null, msg);
    }
    return {
        spawn: spawn,
        destroy: destroy
    };
}
exports.factory = factory;
//# sourceMappingURL=PtyProxy.js.map