"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const crypto = require("crypto");
const _ = require("lodash");
const utf8 = require("utf8");
const ViewerElement_1 = require("./ViewerElement");
const ViewerElementTypes = require("./ViewerElementTypes");
const ResizeRefreshElementBase = require("./ResizeRefreshElementBase");
const ResizeCanary_1 = require("./ResizeCanary");
const ThemeableElementBase_1 = require("./ThemeableElementBase");
const ThemeTypes = require("./Theme");
const EmbeddedViewer_1 = require("./EmbeddedViewer");
const CommandPlaceholder_1 = require("./CommandPlaceholder");
const TerminalViewer_1 = require("./viewers/TerminalViewer");
const TextViewer_1 = require("./viewers/TextViewer");
const ImageViewer_1 = require("./viewers/ImageViewer");
const TipViewer_1 = require("./viewers/TipViewer");
const GeneralEvents = require("./GeneralEvents");
const KeyBindingManager_1 = require("./KeyBindingManager");
const CommandPaletteRequestTypes_1 = require("./CommandPaletteRequestTypes");
const Logger_1 = require("./Logger");
const LogDecorator_1 = require("./LogDecorator");
const DomUtils = require("./DomUtils");
const Term = require("./emulator/Term");
const ScrollBar_1 = require("./gui/ScrollBar");
const WebIpc = require("./WebIpc");
const VirtualScrollArea = require("./VirtualScrollArea");
const MmeTypeDetector = require("./MimeTypeDetector");
const CodeMirrorOperation = require("./CodeMirrorOperation");
const Config_1 = require("./Config");
const SupportsClipboardPaste = require("./SupportsClipboardPaste");
const VisualState = ViewerElementTypes.VisualState;
const Mode = ViewerElementTypes.Mode; // This gets us access to the object holding the enum values.
const log = LogDecorator_1.default;
const DEBUG = true;
const DEBUG_APPLICATION_MODE = false;
let startTime = window.performance.now();
let registered = false;
const ID = "EtTerminalTemplate";
exports.EXTRATERM_COOKIE_ENV = "LC_EXTRATERM_COOKIE";
const ID_SCROLL_CONTAINER = "ID_SCROLL_CONTAINER";
const ID_SCROLL_AREA = "ID_SCROLL_AREA";
const ID_SCROLLBAR = "ID_SCROLLBAR";
const ID_CONTAINER = "ID_CONTAINER";
const ID_CSS_VARS = "ID_CSS_VARS";
const KEYBINDINGS_DEFAULT_MODE = "terminal-default-mode";
const KEYBINDINGS_CURSOR_MODE = "terminal-cursor-mode";
const PALETTE_GROUP = "terminal";
const COMMAND_ENTER_CURSOR_MODE = "enterCursorMode";
const COMMAND_ENTER_NORMAL_MODE = "enterNormalMode";
const COMMAND_SCROLL_PAGE_UP = "scrollPageUp";
const COMMAND_SCROLL_PAGE_DOWN = "scrollPageDown";
const COMMAND_COPY_TO_CLIPBOARD = "copyToClipboard";
const COMMAND_PASTE_FROM_CLIPBOARD = "pasteFromClipboard";
const COMMAND_DELETE_LAST_FRAME = "deleteLastFrame";
const COMMAND_OPEN_LAST_FRAME = "openLastFrame";
const COMMAND_RESET_VT = "resetVT";
const COMMAND_CLEAR_SCROLLBACK = "clearScrollback";
const COMMAND_FONT_SIZE_INCREASE = "increaseFontSize";
const COMMAND_FONT_SIZE_DECREASE = "decreaseFontSize";
const COMMAND_FONT_SIZE_RESET = "resetFontSize";
const COMMAND_GO_TO_PREVIOUS_FRAME = "goToPreviousFrame";
const COMMAND_GO_TO_NEXT_FRAME = "goToNextFrame";
const CHILD_RESIZE_BATCH_SIZE = 3;
const CLASS_CURSOR_MODE = "cursor-mode";
const SCROLL_STEP = 1;
const MILLIS_PER_DAY = 1000 * 60 * 60 * 24;
const MINIMUM_FONT_SIZE = -3;
const MAXIMUM_FONT_SIZE = 4;
// List of viewer classes.
const viewerClasses = [];
viewerClasses.push(ImageViewer_1.ImageViewer);
viewerClasses.push(TextViewer_1.TextViewer);
viewerClasses.push(TipViewer_1.TipViewer);
/**
 * An Extraterm terminal.
 *
 * An EtTerminal is full terminal emulator with GUI intergration. It handles the
 * UI chrome wrapped around the smaller terminal emulation part (term.js).
 */
class EtTerminal extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        // The current size of the emulator. This is used to detect changes in size.
        this._columns = -1;
        this._rows = -1;
        this._initProperties();
        this._fetchNextTag();
    }
    /**
     * Initialize the EtTerminal class and resources.
     *
     * When EtTerminal is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            ScrollBar_1.ScrollBar.init();
            EmbeddedViewer_1.EmbeddedViewer.init();
            CommandPlaceholder_1.CommandPlaceHolder.init();
            TerminalViewer_1.TerminalViewer.init();
            TextViewer_1.TextViewer.init();
            ImageViewer_1.ImageViewer.init();
            TipViewer_1.TipViewer.init();
            ResizeCanary_1.ResizeCanary.init();
            // EtMarkdownViewer.init();
            window.customElements.define(EtTerminal.TAG_NAME.toLowerCase(), EtTerminal);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger(EtTerminal.TAG_NAME, this);
        this._virtualScrollArea = null;
        this._stashArea = null;
        this._childElementList = [];
        this._elementAttached = false;
        this._needsCompleteRefresh = true;
        this._autoscroll = true;
        this._emulator = null;
        this._cookie = null;
        this._terminalViewer = null;
        this._htmlData = null;
        this._showPayloadSize = null;
        this._showData = null;
        this._applicationMode = 0 /* APPLICATION_MODE_NONE */;
        this._bracketStyle = null;
        this._mode = Mode.DEFAULT;
        this._configManager = null;
        this._keyBindingManager = null;
        this._frameFinder = null;
        this._title = "New Tab";
        this._nextTag = null;
        this._themeCssPath = null;
        this._mainStyleLoaded = false;
        this._themeStyleLoaded = false;
        this._resizePollHandle = null;
        this._enforceScrollbackLengthGuard = false;
        this._scheduleLaterHandle = null;
        this._scheduleLaterQueue = [];
        this._stashedChildResizeTask = null;
        this._fontSizeAdjustment = 0;
        this._armResizeCanary = false;
        this._lastCommandLine = null;
        this._lastCommandTerminalViewer = null;
        this._lastCommandTerminalLine = null;
        this._childFocusHandlerFunc = this._handleChildFocus.bind(this);
    }
    //-----------------------------------------------------------------------
    //
    //   ######                                
    //   #     # #    # #####  #      #  ####  
    //   #     # #    # #    # #      # #    # 
    //   ######  #    # #####  #      # #      
    //   #       #    # #    # #      # #      
    //   #       #    # #    # #      # #    # 
    //   #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    /**
     * Blinking cursor
     *
     * True means the cursor should blink, otherwise it doesn't.
     */
    setBlinkingCursor(blink) {
        // this._blinkingCursor = blink;
        // if (this._term !== null) {
        //   this._term.setCursorBlink(blink);
        // }
    }
    /**
     * The number of columns in the terminal screen.
     */
    getColumns() {
        return this._columns;
    }
    /**
     * The number of rows in the terminal screen.
     */
    getRows() {
        return this._rows;
    }
    setConfigManager(configManager) {
        this._configManager = configManager;
    }
    setKeyBindingManager(keyBindingManager) {
        this._keyBindingManager = keyBindingManager;
    }
    setScrollbackSize(scrollbackSize) {
        this._scrollbackSize = scrollbackSize;
    }
    getScrollbackSize() {
        return this._scrollbackSize;
    }
    _isNoFrameCommand(commandLine) {
        const cleanCommandLine = commandLine.trim();
        if (cleanCommandLine === "") {
            return true;
        }
        const commandParts = cleanCommandLine.split(/\s+/);
        if (this._configManager === null) {
            return false;
        }
        else {
            const commandLineActions = this._configManager.getConfig().commandLineActions || [];
            return commandLineActions
                .filter(cla => !cla.frame)
                .some(cla => {
                if (cla.matchType === 'name') {
                    const matcherParts = cla.match.split(/\s+/);
                    for (let i = 0; i < matcherParts.length; i++) {
                        if (i >= commandParts.length) {
                            return false;
                        }
                        if (matcherParts[i] !== commandParts[i]) {
                            return false;
                        }
                    }
                    return true;
                }
                else {
                    // regexp
                    return (new RegExp(cla.match)).test(cleanCommandLine);
                }
            });
        }
    }
    /**
     * Get this terminal's title.
     *
     * This is the window title of the terminal, don't confuse it with more
     * general HTML title of the element.
     */
    getTerminalTitle() {
        return this._title;
    }
    /**
     * Destroy the terminal.
     */
    destroy() {
        if (this._resizePollHandle !== null) {
            this._resizePollHandle.cancel();
            this._resizePollHandle = null;
        }
        this._getWindow().removeEventListener('resize', this._scheduleResizeBound);
        if (this._emulator !== null) {
            this._emulator.destroy();
        }
        this._emulator = null;
    }
    /**
     * Focus on this terminal.
     */
    focus() {
        if (this._terminalViewer !== null) {
            DomUtils.focusWithoutScroll(this._terminalViewer);
        }
    }
    /**
     * Returns true if this terminal has the input focus.
     *
     * @return true if the terminal has the focus.
     */
    hasFocus() {
        const shadowRoot = DomUtils.getShadowRoot(this);
        if (shadowRoot === null) {
            return false;
        }
        return shadowRoot.activeElement !== null;
    }
    /**
     * Write data to the terminal screen.
     *
     * @param text the stream of data to write.
     */
    write(text) {
        return this._emulator.write(text);
    }
    /**
     * Send data to the pty and process connected to the terminal.
     * @param text the data to send.
     */
    send(text) {
        this._sendDataToPtyEvent(text);
    }
    resizeToContainer() {
        this._scheduleResize();
    }
    setFrameFinder(func) {
        this._frameFinder = func;
    }
    getFrameContents(frameId) {
        const embeddedViewer = this.getEmbeddedViewerByFrameId(frameId);
        if (embeddedViewer === null) {
            return null;
        }
        const text = embeddedViewer.getText();
        return text === undefined ? null : text;
    }
    getEmbeddedViewerByFrameId(frameId) {
        if (/[^0-9]/.test(frameId)) {
            return null;
        }
        for (const elementStat of this._childElementList) {
            const element = elementStat.element;
            if (EmbeddedViewer_1.EmbeddedViewer.is(element) && element.getAttribute('tag') === frameId) {
                return element;
            }
        }
        return null;
    }
    getFontAdjust() {
        return this._fontSizeAdjustment;
    }
    setFontAdjust(delta) {
        this._adjustFontSize(delta);
    }
    getViewerElements() {
        return this._childElementList
            .map(status => status.element)
            .filter(el => el instanceof ViewerElement_1.ViewerElement);
    }
    getExtratermCookieValue() {
        return this._cookie;
    }
    /**
     * Custom Element 'connected' life cycle hook.
     */
    connectedCallback() {
        super.connectedCallback();
        if (!this._elementAttached) {
            this._elementAttached = true;
            this._stashArea = window.document.createDocumentFragment();
            this._stashArea.addEventListener(VirtualScrollArea.EVENT_RESIZE, this._handleVirtualScrollableResize.bind(this));
            const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
            const clone = this._createClone();
            shadow.appendChild(clone);
            this.addEventListener('focus', this._handleFocus.bind(this));
            this.addEventListener('blur', this._handleBlur.bind(this));
            const scrollbar = DomUtils.getShadowId(this, ID_SCROLLBAR);
            const scrollArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
            const scrollContainer = DomUtils.getShadowId(this, ID_SCROLL_CONTAINER);
            DomUtils.preventScroll(scrollContainer);
            DomUtils.addCustomEventResender(scrollContainer, GeneralEvents.EVENT_DRAG_STARTED, this);
            DomUtils.addCustomEventResender(scrollContainer, GeneralEvents.EVENT_DRAG_ENDED, this);
            this._virtualScrollArea = new VirtualScrollArea.VirtualScrollArea();
            this._virtualScrollArea.setScrollFunction((offset) => {
                scrollArea.style.top = "-" + offset + "px";
            });
            this._virtualScrollArea.setScrollbar(scrollbar);
            this._virtualScrollArea.setSetTopFunction(this._setTopFunction.bind(this));
            this._virtualScrollArea.setMarkVisibleFunction(this._markVisible.bind(this));
            // Set up the emulator
            this._cookie = crypto.randomBytes(10).toString('hex');
            process.env[exports.EXTRATERM_COOKIE_ENV] = this._cookie;
            this._initEmulator(this._cookie);
            this._appendNewTerminalViewer();
            this.updateThemeCss();
            scrollContainer.addEventListener('mousedown', (ev) => {
                if (ev.target === scrollContainer) {
                    ev.preventDefault();
                    ev.stopPropagation();
                    this._terminalViewer.focus();
                    if (ev.buttons & 2) {
                        this._handleContextMenu();
                    }
                }
            });
            scrollArea.addEventListener('mousedown', (ev) => {
                if (ev.target === scrollArea) {
                    this._terminalViewer.focus();
                    ev.preventDefault();
                    ev.stopPropagation();
                }
            });
            scrollbar.addEventListener('scroll', (ev) => {
                this._virtualScrollArea.scrollTo(scrollbar.getPosition());
            });
            scrollArea.addEventListener('wheel', this._handleMouseWheel.bind(this), true);
            scrollContainer.addEventListener('mousedown', this._handleMouseDown.bind(this), true);
            scrollArea.addEventListener('keydown', this._handleKeyDownCapture.bind(this), true);
            scrollArea.addEventListener(VirtualScrollArea.EVENT_RESIZE, this._handleVirtualScrollableResize.bind(this));
            scrollArea.addEventListener(TerminalViewer_1.TerminalViewer.EVENT_KEYBOARD_ACTIVITY, () => {
                this._virtualScrollArea.scrollToBottom();
            });
            scrollArea.addEventListener(ViewerElement_1.ViewerElement.EVENT_BEFORE_SELECTION_CHANGE, this._handleBeforeSelectionChange.bind(this));
            scrollArea.addEventListener(ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE, this._handleTerminalViewerCursor.bind(this));
            scrollArea.addEventListener(ViewerElement_1.ViewerElement.EVENT_CURSOR_EDGE, this._handleTerminalViewerCursorEdge.bind(this));
            scrollArea.addEventListener(GeneralEvents.EVENT_TYPE_TEXT, (ev) => {
                const detail = ev.detail;
                this._sendDataToPtyEvent(ev.detail.text);
            });
            // A Resize Canary for tracking when terminal fonts are effectively changed in the DOM.
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            const resizeCanary = document.createElement(ResizeCanary_1.ResizeCanary.TAG_NAME);
            resizeCanary.setCss(`
          font-family: var(--terminal-font);
          font-size: var(--terminal-font-size);
      `);
            containerDiv.appendChild(resizeCanary);
            resizeCanary.addEventListener('resize', () => {
                if (this._armResizeCanary) {
                    this._armResizeCanary = false;
                    this.refresh(ResizeRefreshElementBase.RefreshLevel.COMPLETE);
                }
            });
            this._showTip();
            this._scheduleResize();
        }
        else {
            // This was already attached at least once.
            this._scheduleResize();
        }
        this._setFontSizeInCss(this._fontSizeAdjustment);
    }
    /**
     * Custom Element 'disconnected' life cycle hook.
     */
    disconnectedCallback() {
        super.disconnectedCallback();
        this._needsCompleteRefresh = true;
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.TERMINAL];
    }
    updateThemeCss() {
        super.updateThemeCss();
        this.resizeToContainer();
    }
    refresh(level) {
        this._processRefresh(level);
    }
    //-----------------------------------------------------------------------
    //
    //   ######                                      
    //   #     # #####  # #    #   ##   ##### ###### 
    //   #     # #    # # #    #  #  #    #   #      
    //   ######  #    # # #    # #    #   #   #####  
    //   #       #####  # #    # ######   #   #      
    //   #       #   #  #  #  #  #    #   #   #      
    //   #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    _createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
      <style id="${ID_CSS_VARS}">${this._getCssVarsRules()}</style>
        <div id='${ID_CONTAINER}'>
          <div id='${ID_SCROLL_CONTAINER}'>
            <div id='${ID_SCROLL_AREA}'></div>
          </div>
          <${ScrollBar_1.ScrollBar.TAG_NAME} id='${ID_SCROLLBAR}'></${ScrollBar_1.ScrollBar.TAG_NAME}>
        </div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _getCssVarsRules() {
        return `
    #${ID_CONTAINER} {
        ${this._getCssFontSizeRule(this._fontSizeAdjustment)}
    }
    `;
    }
    _getCssFontSizeRule(adjustment) {
        const scale = [0.6, 0.75, 0.89, 1, 1.2, 1.5, 2, 3][adjustment - MINIMUM_FONT_SIZE];
        return `--terminal-font-size: calc(var(--default-terminal-font-size) * ${scale});`;
    }
    /**
     * Get the window which this terminal is on.
     *
     * @returns {Window} The window object.
     */
    _getWindow() {
        return this.ownerDocument.defaultView;
    }
    _getDocument() {
        return this.ownerDocument;
    }
    _showTip() {
        const config = this._configManager.getConfig();
        switch (config.showTips) {
            case 'always':
                break;
            case 'never':
                return;
            case 'daily':
                if ((Date.now() - config.tipTimestamp) < MILLIS_PER_DAY) {
                    return;
                }
        }
        this._appendMimeViewer(TipViewer_1.TipViewer.MIME_TYPE, "Tip", "utf8", null);
        const newConfig = _.cloneDeep(config);
        newConfig.tipTimestamp = Date.now();
        newConfig.tipCounter = newConfig.tipCounter + 1;
        this._configManager.setConfig(newConfig);
    }
    _handleFocus(event) {
        // Forcefully set the visual state of each thing in the terminal to appear focused.
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        DomUtils.nodeListToArray(scrollerArea.childNodes).forEach((node) => {
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                node.setVisualState(this._mode === Mode.CURSOR ? VisualState.AUTO : VisualState.FOCUSED);
            }
        });
    }
    _refocus() {
        if (this.hasFocus() && this._mode === Mode.DEFAULT && this._terminalViewer != null && !this._terminalViewer.hasFocus()) {
            DomUtils.focusWithoutScroll(this._terminalViewer);
        }
    }
    _handleBlur(event) {
        // Forcefully set the visual state of each thing in the terminal to appear unfocused.
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        DomUtils.nodeListToArray(scrollerArea.childNodes).forEach((node) => {
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                node.setVisualState(VisualState.UNFOCUSED);
            }
        });
    }
    _handleBeforeSelectionChange(ev) {
        const target = ev.target;
        this._childElementList.forEach((nodeInfo) => {
            const node = nodeInfo.element;
            if (ViewerElement_1.ViewerElement.isViewerElement(node) && node !== target) {
                node.clearSelection();
            }
        });
        if (ev.detail.originMouse) {
            DomUtils.doLater(() => { this.copyToClipboard(); }); // FIXME This should be debounced slightly.
        }
    }
    _handleChildFocus(ev) {
        // This needs to be done later otherwise it tickles a bug in
        // Chrome/Blink and prevents drag and drop from working.
        // https://bugs.chromium.org/p/chromium/issues/detail?id=726248
        DomUtils.doLater(() => {
            if (this._mode === Mode.DEFAULT) {
                this.focus();
            }
        });
    }
    // ----------------------------------------------------------------------
    //
    // #######                                                 
    // #       #    # #    # #        ##   #####  ####  #####  
    // #       ##  ## #    # #       #  #    #   #    # #    # 
    // #####   # ## # #    # #      #    #   #   #    # #    # 
    // #       #    # #    # #      ######   #   #    # #####  
    // #       #    # #    # #      #    #   #   #    # #   #  
    // ####### #    #  ####  ###### #    #   #    ####  #    # 
    //                                                         
    // ----------------------------------------------------------------------
    _initEmulator(cookie) {
        const emulator = new Term.Emulator({
            userAgent: window.navigator.userAgent,
            applicationModeCookie: cookie,
            debug: true
        });
        emulator.debug = true;
        emulator.addTitleChangeEventListener(this._handleTitle.bind(this));
        emulator.addDataEventListener(this._handleTermData.bind(this));
        emulator.addRenderEventListener(this._handleTermSize.bind(this));
        // Application mode handlers    
        emulator.addApplicationModeStartEventListener(this._handleApplicationModeStart.bind(this));
        emulator.addApplicationModeDataEventListener(this._handleApplicationModeData.bind(this));
        emulator.addApplicationModeEndEventListener(this._handleApplicationModeEnd.bind(this));
        emulator.addWriteBufferSizeEventListener(this._handleWriteBufferSize.bind(this));
        this._emulator = emulator;
    }
    _appendNewTerminalViewer() {
        // Create the TerminalViewer
        const terminalViewer = document.createElement(TerminalViewer_1.TerminalViewer.TAG_NAME);
        KeyBindingManager_1.injectKeyBindingManager(terminalViewer, this._keyBindingManager);
        Config_1.injectConfigManager(terminalViewer, this._configManager);
        terminalViewer.setEmulator(this._emulator);
        this._appendScrollable(terminalViewer);
        terminalViewer.setVisualState(DomUtils.getShadowRoot(this).activeElement !== null
            ? VisualState.FOCUSED
            : VisualState.UNFOCUSED);
        this._terminalViewer = terminalViewer;
        this._emulator.refreshScreen();
    }
    _appendScrollable(el) {
        el.addEventListener('focus', this._childFocusHandlerFunc);
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        this._childElementList.push({ element: el, needsRefresh: false, refreshLevel: ResizeRefreshElementBase.RefreshLevel.RESIZE });
        scrollerArea.appendChild(el);
        this._virtualScrollArea.appendScrollable(el);
    }
    _removeScrollable(el) {
        el.removeEventListener('focus', this._childFocusHandlerFunc);
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        if (el.parentElement === scrollerArea) {
            scrollerArea.removeChild(el);
        }
        else if (el.parentNode === this._stashArea) {
            this._stashArea.removeChild(el);
        }
        const pos = this._childElementListIndexOf(el);
        this._childElementList.splice(pos, 1);
        this._virtualScrollArea.removeScrollable(el);
    }
    /**
     * Handler for window title change events from the pty.
     *
     * @param title The new window title for this terminal.
     */
    _handleTitle(emulator, title) {
        this._title = title;
        this._sendTitleEvent(title);
    }
    _disconnectActiveTerminalViewer() {
        this._moveCursorToFreshLine();
        this._emulator.moveRowsAboveCursorToScrollback();
        this._emulator.flushRenderQueue();
        if (this._terminalViewer !== null) {
            this._terminalViewer.setEmulator(null);
            this._terminalViewer.deleteScreen();
            this._terminalViewer.setUseVPad(false);
            this._virtualScrollArea.updateScrollableSize(this._terminalViewer);
            this._terminalViewer = null;
        }
    }
    _appendScrollableElement(el) {
        this._emulator.moveRowsAboveCursorToScrollback();
        this._emulator.flushRenderQueue();
        let currentTerminalViewer = this._terminalViewer;
        let currentTerminalViewerHadFocus = false;
        if (currentTerminalViewer !== null) {
            currentTerminalViewer.deleteScreen();
            currentTerminalViewerHadFocus = currentTerminalViewer.hasFocus();
            if (currentTerminalViewer.isEmpty()) {
                // Keep this terminal viewer and re-use it later in the new position.
                this._removeScrollable(currentTerminalViewer);
            }
            else {
                // This terminal viewer has stuff in it.
                currentTerminalViewer.setEmulator(null);
                currentTerminalViewer.setUseVPad(false);
                this._virtualScrollArea.updateScrollableSize(currentTerminalViewer);
                this._terminalViewer = null;
                currentTerminalViewer = null;
            }
        }
        this._appendScrollable(el);
        if (currentTerminalViewer !== null) {
            this._appendScrollable(currentTerminalViewer);
            if (currentTerminalViewerHadFocus) {
                currentTerminalViewer.focus();
            }
        }
        else {
            this._appendNewTerminalViewer();
            this._refocus();
        }
    }
    _replaceScrollableElement(oldEl, newEl) {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        scrollerArea.insertBefore(newEl, oldEl);
        scrollerArea.removeChild(oldEl);
        this._childElementList.splice(this._childElementListIndexOf(oldEl), 1);
        this._virtualScrollArea.replaceScrollable(oldEl, newEl);
    }
    _handleMouseDown(ev) {
        if (ev.buttons === 4) {
            ev.stopPropagation();
            ev.preventDefault();
            this._pasteFromClipboard();
        }
    }
    _handleWriteBufferSize(emulator, status) {
        const event = new CustomEvent(EtTerminal.EVENT_TERMINAL_BUFFER_SIZE, { detail: status });
        this.dispatchEvent(event);
    }
    /**
     * Handle new stdout data from the pty.
     *
     * @param {string} data New data.
     */
    _handlePtyStdoutData(data) {
        // log("incoming data:",""+data);
        this._emulator.write("" + data);
    }
    /**
     * Handle new stderr data from the pty.
     *
     * @param {type} data New data.
     */
    _handlePtyStderrData(data) {
        this._emulator.write(data);
    }
    /**
     * Handle data coming from the user.
     *
     * This just pushes the keys from the user through to the pty.
     * @param {string} data The data to process.
     */
    _handleTermData(emulator, data) {
        this._sendDataToPtyEvent(data);
    }
    _handleTermSize(emulator, event) {
        const newColumns = event.columns;
        const newRows = event.rows;
        if (this._columns === newColumns && this._rows === newRows) {
            return;
        }
        this._columns = newColumns;
        this._rows = newRows;
        this._sendResizeEvent(newColumns, newRows);
    }
    /**
     * Send data to the pseudoterminal.
     *
     * @param {string} text
     */
    _sendDataToPtyEvent(text) {
        const event = new CustomEvent(EtTerminal.EVENT_USER_INPUT, { detail: { data: text } });
        this.dispatchEvent(event);
    }
    /**
     * Send a resize message to the pty.
     *
     * @param {number} cols The new number of columns in the terminal.
     * @param {number} rows The new number of rows in the terminal.
     */
    _sendResizeEvent(cols, rows, callback) {
        const event = new CustomEvent(EtTerminal.EVENT_TERMINAL_RESIZE, { detail: { columns: cols, rows: rows } });
        this.dispatchEvent(event);
    }
    _sendTitleEvent(title) {
        const event = new CustomEvent(EtTerminal.EVENT_TITLE, { detail: { title: title } });
        this.dispatchEvent(event);
    }
    _enterCursorMode() {
        this._setModeAndVisualState(ViewerElementTypes.Mode.CURSOR, VisualState.AUTO);
        this._mode = Mode.CURSOR;
    }
    _exitCursorMode() {
        this._setModeAndVisualState(ViewerElementTypes.Mode.DEFAULT, VisualState.FOCUSED);
        this._mode = Mode.DEFAULT;
        this._refocus();
    }
    _setModeAndVisualState(mode, visualState) {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        const childNodes = DomUtils.nodeListToArray(scrollerArea.childNodes).filter(ViewerElement_1.ViewerElement.isViewerElement);
        childNodes.forEach((node) => node.setMode(mode));
        childNodes.forEach((node) => node.setVisualState(visualState));
    }
    _childElementListIndexOf(element) {
        const list = this._childElementList;
        ;
        const len = list.length;
        for (let i = 0; i < len; i++) {
            const item = list[i];
            if (item.element === element) {
                return i;
            }
        }
        return -1;
    }
    // ----------------------------------------------------------------------
    //
    //    #####                                                          ##        #####                           
    //   #     #  ####  #####   ####  #      #      # #    #  ####      #  #      #     # # ###### # #    #  ####  
    //   #       #    # #    # #    # #      #      # ##   # #    #      ##       #       #     #  # ##   # #    # 
    //    #####  #      #    # #    # #      #      # # #  # #          ###        #####  #    #   # # #  # #      
    //         # #      #####  #    # #      #      # #  # # #  ###    #   # #          # #   #    # #  # # #  ### 
    //   #     # #    # #   #  #    # #      #      # #   ## #    #    #    #     #     # #  #     # #   ## #    # 
    //    #####   ####  #    #  ####  ###### ###### # #    #  ####      ###  #     #####  # ###### # #    #  ####  
    //
    // ----------------------------------------------------------------------
    _handleMouseWheel(ev) {
        ev.stopPropagation();
        ev.preventDefault();
        const delta = ev.deltaY * SCROLL_STEP;
        this._virtualScrollArea.scrollTo(this._virtualScrollArea.getScrollYOffset() + delta);
    }
    _handleVirtualScrollableResize(ev) {
        const el = ev.target;
        if (el.parentNode === this._stashArea) {
            this._scheduleStashedChildResize(el);
        }
        else {
            this._updateVirtualScrollableSize(el);
        }
    }
    _markVisible(scrollable, visible) {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        const element = scrollable;
        if (!visible) {
            if (this._terminalViewer !== element && !(ViewerElement_1.ViewerElement.isViewerElement(element) && element.hasFocus())) {
                // Move the scrollable into the stash area.
                this._stashArea.appendChild(element);
            }
        }
        else {
            if (element.parentElement !== scrollerArea) {
                // Move the element to the scroll area and place it in the correct position relative to the other child elements.
                const scrollerAreaChildrenCount = scrollerArea.children.length;
                if (scrollerAreaChildrenCount === 0) {
                    scrollerArea.appendChild(element);
                }
                else {
                    let scrollerIndex = 0;
                    let childIndex = 0;
                    while (childIndex < this._childElementList.length) {
                        const currentElement = this._childElementList[childIndex].element;
                        if (currentElement === element) {
                            scrollerArea.insertBefore(element, scrollerArea.children[scrollerIndex]);
                            break;
                        }
                        if (scrollerArea.children[scrollerIndex] === currentElement) {
                            scrollerIndex++;
                            if (scrollerIndex >= scrollerAreaChildrenCount) {
                                scrollerArea.appendChild(element);
                                break;
                            }
                        }
                        childIndex++;
                    }
                }
                // Set the current mode on the scrollable.
                const visualState = this._mode === Mode.CURSOR ? VisualState.AUTO : VisualState.FOCUSED;
                element.setMode(this._mode);
                element.setVisualState(visualState);
            }
        }
    }
    _makeVisible(element) {
        this._markVisible(element, true);
    }
    _updateVirtualScrollableSize(virtualScrollable) {
        this._virtualScrollArea.updateScrollableSize(virtualScrollable);
        this._enforceScrollbackLength(this._scrollbackSize);
    }
    _processRefresh(requestedLevel) {
        let level = requestedLevel;
        if (this._needsCompleteRefresh) {
            level = ResizeRefreshElementBase.RefreshLevel.COMPLETE;
            this._needsCompleteRefresh = false;
        }
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        if (scrollerArea !== null) {
            // --- DOM Read ---
            CodeMirrorOperation.bulkOperation(() => {
                ResizeRefreshElementBase.ResizeRefreshElementBase.refreshChildNodes(scrollerArea, level);
                const scrollbar = DomUtils.getShadowId(this, ID_SCROLLBAR);
                scrollbar.refresh(level);
                // --- DOM write ---
                const scrollContainer = DomUtils.getShadowId(this, ID_SCROLL_CONTAINER);
                this._virtualScrollArea.updateContainerHeight(scrollContainer.getBoundingClientRect().height);
                // Build the list of elements we will resize right now.
                const childrenToResize = [];
                const len = scrollerArea.children.length;
                for (let i = 0; i < len; i++) {
                    const child = scrollerArea.children[i];
                    if (ViewerElement_1.ViewerElement.isViewerElement(child)) {
                        childrenToResize.push(child);
                    }
                }
                // Keep track of which children will need a resize later on.
                const childrenToResizeSet = new Set(childrenToResize);
                for (const childStatus of this._childElementList) {
                    if (!childrenToResizeSet.has(childStatus.element)) {
                        childStatus.needsRefresh = true;
                        childStatus.refreshLevel = level;
                    }
                }
                if (childrenToResize.length !== this._childElementList.length) {
                    this._scheduleStashedChildResizeTask();
                }
                this._virtualScrollArea.updateScrollableSizes(childrenToResize);
                this._virtualScrollArea.reapplyState();
                this._enforceScrollbackLength(this._scrollbackSize);
            });
        }
    }
    _setTopFunction(scrollable, top) {
        scrollable.style.top = "" + top + "px";
    }
    _handleTerminalViewerCursor(ev) {
        const node = ev.target;
        if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
            const pos = node.getCursorPosition();
            const nodeTop = this._virtualScrollArea.getScrollableTop(node);
            const top = pos.top + nodeTop;
            const bottom = pos.bottom + nodeTop;
            this._virtualScrollArea.scrollIntoView(top, bottom);
        }
        else {
            this._log.warn("_handleTerminalViewerCursor(): node is not a ViewerElement.");
        }
    }
    _handleTerminalViewerCursorEdge(ev) {
        const detail = ev.detail;
        const index = this._childElementListIndexOf(ev.target);
        if (index === -1) {
            this._log.warn("_handleTerminalViewerCursorEdge(): Couldn't find the target.");
            return;
        }
        if (detail.edge === ViewerElementTypes.Edge.TOP) {
            // A top edge was hit. Move the cursor to the bottom of the ViewerElement above it.
            for (let i = index - 1; i >= 0; i--) {
                const node = this._childElementList[i].element;
                if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                    this._makeVisible(node);
                    if (node.setCursorPositionBottom(detail.ch)) {
                        DomUtils.focusWithoutScroll(node);
                        break;
                    }
                }
            }
        }
        else {
            // Bottom edge. Move the cursor to the top of the next ViewerElement.
            for (let i = index + 1; i < this._childElementList.length; i++) {
                const node = this._childElementList[i].element;
                if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                    this._makeVisible(node);
                    if (node.setCursorPositionTop(detail.ch)) {
                        DomUtils.focusWithoutScroll(node);
                        break;
                    }
                }
            }
        }
    }
    // Run a function and only afterwards check the size of the scrollback.
    _enforceScrollbackLengthAfter(func) {
        const oldGuardFlag = this._enforceScrollbackLengthGuard;
        this._enforceScrollbackLengthGuard = true;
        const rc = func();
        this._enforceScrollbackLengthGuard = oldGuardFlag;
        this._enforceScrollbackLength(this._scrollbackSize);
        return rc;
    }
    _enforceScrollbackLength(scrollbackSize) {
        // Prevent the scrollback check from running multiple times.
        if (this._enforceScrollbackLengthGuard) {
            return;
        }
        this._enforceScrollbackLengthGuard = true;
        const hasFocus = this.hasFocus();
        this._enforceScrollbackLength2(scrollbackSize);
        if (hasFocus && !this.hasFocus()) {
            this.focus();
        }
        this._enforceScrollbackLengthGuard = false;
    }
    _enforceScrollbackLength2(scrollbackSize) {
        let virtualHeight = this._virtualScrollArea.getVirtualHeight();
        const effectiveScrollbackSize = window.screen.height + scrollbackSize;
        const hardLimit = Math.floor(effectiveScrollbackSize * 1.1);
        if (virtualHeight < hardLimit) {
            return;
        }
        const killList = [];
        for (const nodeInfo of this._childElementList) {
            const scrollableKid = nodeInfo.element;
            const kidVirtualHeight = this._virtualScrollArea.getScrollableVirtualHeight(scrollableKid);
            const newVirtualHeight = virtualHeight - kidVirtualHeight;
            // We don't want to cut out too much at once.
            if (newVirtualHeight > effectiveScrollbackSize) {
                // Just remove the thing. There is plenty of scrollback left over.
                killList.push(scrollableKid);
            }
            else {
                this._deleteTopPixels(scrollableKid, virtualHeight - effectiveScrollbackSize);
                break;
            }
            virtualHeight = newVirtualHeight;
            if (virtualHeight < hardLimit) {
                break;
            }
        }
        for (const scrollableKid of killList) {
            this._removeScrollable(scrollableKid);
        }
    }
    _deleteTopPixels(kidNode, pixelCount) {
        // Try to cut part of it off.
        if (TerminalViewer_1.TerminalViewer.is(kidNode)) {
            kidNode.deleteTopPixels(pixelCount);
            this._scheduleStashedChildResize(kidNode);
            return;
        }
        else if (EmbeddedViewer_1.EmbeddedViewer.is(kidNode)) {
            const viewer = kidNode.getViewerElement();
            if (TerminalViewer_1.TerminalViewer.is(viewer)) {
                viewer.deleteTopPixels(pixelCount);
                this._scheduleStashedChildResize(kidNode);
                return;
            }
            else if (TextViewer_1.TextViewer.is(viewer)) {
                viewer.deleteTopPixels(pixelCount);
                this._scheduleStashedChildResize(kidNode);
                return;
            }
        }
        this._removeScrollable(kidNode);
    }
    _adjustFontSize(delta) {
        const newAdjustment = Math.min(Math.max(this._fontSizeAdjustment + delta, MINIMUM_FONT_SIZE), MAXIMUM_FONT_SIZE);
        if (newAdjustment !== this._fontSizeAdjustment) {
            this._fontSizeAdjustment = newAdjustment;
            this._setFontSizeInCss(newAdjustment);
        }
    }
    _setFontSizeInCss(size) {
        const styleElement = DomUtils.getShadowId(this, ID_CSS_VARS);
        styleElement.sheet.cssRules[0].style.cssText = this._getCssFontSizeRule(size); // Type stubs are missing cssRules.
        this._armResizeCanary = true;
        // Don't refresh. Let the Resize Canary detect the real change in the DOM when it arrives.
    }
    _resetFontSize() {
        this._adjustFontSize(-this._fontSizeAdjustment);
    }
    _goToPreviousFrame() {
        const heights = this._virtualScrollArea.getScrollableHeights();
        const y = this._virtualScrollArea.getScrollYOffset();
        let heightCount = 0;
        for (let i = 0; i < heights.length; i++) {
            if (y <= (heightCount + heights[i].height)) {
                this._virtualScrollArea.scrollTo(heightCount);
                break;
            }
            heightCount += heights[i].height;
        }
    }
    _goToNextFrame() {
        const heights = this._virtualScrollArea.getScrollableHeights();
        const y = this._virtualScrollArea.getScrollYOffset();
        let heightCount = 0;
        for (let i = 0; i < heights.length; i++) {
            if (y < (heightCount + heights[i].height)) {
                this._virtualScrollArea.scrollTo(heightCount + heights[i].height);
                break;
            }
            heightCount += heights[i].height;
        }
    }
    // ----------------------------------------------------------------------
    //
    //   #    #                                                 
    //   #   #  ###### #   # #####   ####    ##   #####  #####  
    //   #  #   #       # #  #    # #    #  #  #  #    # #    # 
    //   ###    #####    #   #####  #    # #    # #    # #    # 
    //   #  #   #        #   #    # #    # ###### #####  #    # 
    //   #   #  #        #   #    # #    # #    # #   #  #    # 
    //   #    # ######   #   #####   ####  #    # #    # #####  
    //                                                        
    // ----------------------------------------------------------------------
    _handleKeyDownCapture(ev) {
        if (this._terminalViewer === null || this._keyBindingManager === null ||
            this._keyBindingManager.getKeyBindingContexts() === null) {
            return;
        }
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(this._mode === Mode.DEFAULT
            ? KEYBINDINGS_DEFAULT_MODE : KEYBINDINGS_CURSOR_MODE);
        const command = keyBindings.mapEventToCommand(ev);
        if (this._executeCommand(command)) {
            ev.stopPropagation();
            ev.preventDefault();
        }
    }
    _handleContextMenu() {
        if (this._terminalViewer !== null) {
            this._terminalViewer.executeCommand(CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE);
        }
    }
    getCommandPaletteEntries(commandableStack) {
        const commandList = [];
        if (this._mode === Mode.DEFAULT) {
            commandList.push({ id: COMMAND_ENTER_CURSOR_MODE, group: PALETTE_GROUP, iconRight: "i-cursor", label: "Enter cursor mode", commandExecutor: this });
        }
        else {
            commandList.push({ id: COMMAND_ENTER_NORMAL_MODE, group: PALETTE_GROUP, label: "Enter normal mode", commandExecutor: this });
        }
        commandList.push({ id: COMMAND_SCROLL_PAGE_UP, group: PALETTE_GROUP, iconRight: "angle-double-up", label: "Scroll Page Up", commandExecutor: this });
        commandList.push({ id: COMMAND_SCROLL_PAGE_DOWN, group: PALETTE_GROUP, iconRight: "angle-double-down", label: "Scroll Page Down", commandExecutor: this });
        commandList.push({ id: COMMAND_GO_TO_PREVIOUS_FRAME, group: PALETTE_GROUP, label: "Go to Previous Frame", commandExecutor: this });
        commandList.push({ id: COMMAND_GO_TO_NEXT_FRAME, group: PALETTE_GROUP, label: "Go to Next Frame", commandExecutor: this });
        commandList.push({ id: COMMAND_COPY_TO_CLIPBOARD, group: PALETTE_GROUP, iconRight: "copy", label: "Copy to Clipboard", commandExecutor: this });
        if (this._mode === Mode.CURSOR) {
            commandList.push({ id: COMMAND_PASTE_FROM_CLIPBOARD, group: PALETTE_GROUP, iconRight: "clipboard", label: "Paste from Clipboard", commandExecutor: this });
        }
        commandList.push({ id: COMMAND_OPEN_LAST_FRAME, group: PALETTE_GROUP, iconRight: "external-link", label: "Open Last Frame", commandExecutor: this });
        commandList.push({ id: COMMAND_DELETE_LAST_FRAME, group: PALETTE_GROUP, iconRight: "times-circle", label: "Delete Last Frame", commandExecutor: this });
        commandList.push({ id: COMMAND_FONT_SIZE_INCREASE, group: PALETTE_GROUP, label: "Increase Font Size", commandExecutor: this });
        commandList.push({ id: COMMAND_FONT_SIZE_DECREASE, group: PALETTE_GROUP, label: "Decrease Font Size", commandExecutor: this });
        commandList.push({ id: COMMAND_FONT_SIZE_RESET, group: PALETTE_GROUP, label: "Reset Font Size", commandExecutor: this });
        commandList.push({ id: COMMAND_CLEAR_SCROLLBACK, group: PALETTE_GROUP, iconRight: "eraser", label: "Clear Scrollback", commandExecutor: this });
        commandList.push({ id: COMMAND_RESET_VT, group: PALETTE_GROUP, iconRight: "refresh", label: "Reset VT", commandExecutor: this });
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(this._mode === Mode.DEFAULT
            ? KEYBINDINGS_DEFAULT_MODE : KEYBINDINGS_CURSOR_MODE);
        if (keyBindings !== null) {
            commandList.forEach((commandEntry) => {
                const shortcut = keyBindings.mapCommandToKeyBinding(commandEntry.id);
                commandEntry.shortcut = shortcut === null ? "" : shortcut;
            });
        }
        return commandList;
    }
    executeCommand(commandId) {
        this._executeCommand(commandId);
    }
    _executeCommand(command) {
        switch (command) {
            case COMMAND_ENTER_CURSOR_MODE:
                this._enterCursorMode();
                break;
            case COMMAND_ENTER_NORMAL_MODE:
                this._exitCursorMode();
                break;
            case COMMAND_SCROLL_PAGE_UP:
                this._virtualScrollArea.scrollTo(this._virtualScrollArea.getScrollYOffset()
                    - this._virtualScrollArea.getScrollContainerHeight() / 2);
                break;
            case COMMAND_SCROLL_PAGE_DOWN:
                this._virtualScrollArea.scrollTo(this._virtualScrollArea.getScrollYOffset()
                    + this._virtualScrollArea.getScrollContainerHeight() / 2);
                break;
            case COMMAND_GO_TO_PREVIOUS_FRAME:
                this._goToPreviousFrame();
                break;
            case COMMAND_GO_TO_NEXT_FRAME:
                this._goToNextFrame();
                break;
            case COMMAND_COPY_TO_CLIPBOARD:
                this.copyToClipboard();
                break;
            case COMMAND_PASTE_FROM_CLIPBOARD:
                this._pasteFromClipboard();
                break;
            case COMMAND_DELETE_LAST_FRAME:
                this._deleteLastEmbeddedViewer();
                break;
            case COMMAND_OPEN_LAST_FRAME:
                const viewer = this._getLastEmbeddedViewer();
                if (viewer !== null) {
                    this._embeddedViewerPopOutEvent(viewer);
                }
                break;
            case COMMAND_RESET_VT:
                this._emulator.reset();
                break;
            case COMMAND_CLEAR_SCROLLBACK:
                this._enforceScrollbackLength(0);
                break;
            case COMMAND_FONT_SIZE_INCREASE:
                this._adjustFontSize(1);
                break;
            case COMMAND_FONT_SIZE_DECREASE:
                this._adjustFontSize(-1);
                break;
            case COMMAND_FONT_SIZE_RESET:
                this._resetFontSize();
                break;
            default:
                return false;
        }
        return true;
    }
    // ********************************************************************
    //
    //    #####                                                            
    //   #     #  ####  #    # ###### #####  #    # #      # #    #  ####  
    //   #       #    # #    # #      #    # #    # #      # ##   # #    # 
    //    #####  #      ###### #####  #    # #    # #      # # #  # #      
    //         # #      #    # #      #    # #    # #      # #  # # #  ### 
    //   #     # #    # #    # #      #    # #    # #      # #   ## #    # 
    //    #####   ####  #    # ###### #####   ####  ###### # #    #  ####  
    //
    // ********************************************************************
    _scheduleResize() {
        this._scheduleLaterProcessing(() => {
            this._processRefresh(ResizeRefreshElementBase.RefreshLevel.RESIZE);
        });
    }
    _scheduleStashedChildResize(el) {
        if (el.parentNode !== this._stashArea) {
            return;
        }
        for (const childInfo of this._childElementList) {
            if (childInfo.element === el) {
                if (!childInfo.needsRefresh) {
                    childInfo.needsRefresh = true;
                    childInfo.refreshLevel = ResizeRefreshElementBase.RefreshLevel.RESIZE;
                    this._scheduleStashedChildResizeTask();
                }
                return;
            }
        }
        this._log.warn("_scheduleStashedChildResize() called with an unknown element instance.");
    }
    _scheduleStashedChildResizeTask() {
        if (this._stashedChildResizeTask == null) {
            this._stashedChildResizeTask = () => {
                // Gather the list of elements/scrollables that need refreshing and updating.
                const processList = [];
                for (let i = this._childElementList.length - 1; i >= 0 && processList.length < CHILD_RESIZE_BATCH_SIZE; i--) {
                    const childStatus = this._childElementList[i];
                    if (childStatus.needsRefresh) {
                        processList.push(childStatus);
                        childStatus.needsRefresh = false;
                    }
                }
                if (processList.length !== 0) {
                    // Find the elements which need to be moved into the scroll area.
                    const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
                    const stashedList = [];
                    for (const childStatus of processList) {
                        const element = childStatus.element;
                        if (element.parentElement !== scrollerArea) {
                            stashedList.push(element);
                        }
                    }
                    CodeMirrorOperation.bulkOperation(() => {
                        stashedList.forEach(el => this._markVisible(el, true));
                        for (const childStatus of processList) {
                            const el = childStatus.element;
                            if (ResizeRefreshElementBase.ResizeRefreshElementBase.is(el)) {
                                el.refresh(childStatus.refreshLevel);
                            }
                        }
                        this._virtualScrollArea.updateScrollableSizes(processList.map(childStatus => childStatus.element));
                    });
                    if (stashedList.length !== 0) {
                        stashedList.filter((el) => !this._virtualScrollArea.getScrollableVisible(el))
                            .forEach((el) => this._markVisible(el, false));
                    }
                    this._scheduleStashedChildResizeTask();
                }
            };
        }
        if (this._scheduleLaterQueue.indexOf(this._stashedChildResizeTask) === -1) {
            this._scheduleLaterProcessing(this._stashedChildResizeTask);
        }
    }
    _scheduleLaterProcessing(func) {
        this._scheduleLaterQueue.push(func);
        if (this._scheduleLaterHandle === null) {
            this._scheduleLaterHandle = DomUtils.doLater(() => {
                this._scheduleLaterHandle = null;
                const queue = this._scheduleLaterQueue;
                this._scheduleLaterQueue = [];
                queue.forEach((func) => func());
            });
        }
    }
    // ********************************************************************
    //
    //      #                                                                  #     #                      
    //     # #   #####  #####  #      #  ####    ##   ##### #  ####  #    #    ##   ##  ####  #####  ###### 
    //    #   #  #    # #    # #      # #    #  #  #    #   # #    # ##   #    # # # # #    # #    # #      
    //   #     # #    # #    # #      # #      #    #   #   # #    # # #  #    #  #  # #    # #    # #####  
    //   ####### #####  #####  #      # #      ######   #   # #    # #  # #    #     # #    # #    # #      
    //   #     # #      #      #      # #    # #    #   #   # #    # #   ##    #     # #    # #    # #      
    //   #     # #      #      ###### #  ####  #    #   #   #  ####  #    #    #     #  ####  #####  ###### 
    //
    // ********************************************************************
    /**
     * Handle when the embedded term.js enters start of application mode.
     *
     * @param {array} params The list of parameter which were specified in the
     *     escape sequence.
     */
    _handleApplicationModeStart(emulator, params) {
        if (DEBUG_APPLICATION_MODE) {
            this._log.debug("application-mode started! ", params);
        }
        this._htmlData = "";
        // Check security cookie
        if (params.length === 0) {
            this._log.warn("Received an application mode sequence with no parameters.");
            return;
        }
        if (params[0] !== this._cookie) {
            this._log.warn("Received the wrong cookie at the start of an application mode sequence.");
            return;
        }
        if (params.length === 1) {
            // Normal HTML mode.
            this._applicationMode = 1 /* APPLICATION_MODE_HTML */;
        }
        else if (params.length >= 2) {
            switch ("" + params[1]) {
                case "" + 2 /* APPLICATION_MODE_OUTPUT_BRACKET_START */:
                    this._applicationMode = 2 /* APPLICATION_MODE_OUTPUT_BRACKET_START */;
                    this._bracketStyle = params[2];
                    break;
                case "" + 3 /* APPLICATION_MODE_OUTPUT_BRACKET_END */:
                    this._applicationMode = 3 /* APPLICATION_MODE_OUTPUT_BRACKET_END */;
                    if (DEBUG_APPLICATION_MODE) {
                        this._log.debug("Starting APPLICATION_MODE_OUTPUT_BRACKET_END");
                    }
                    break;
                case "" + 4 /* APPLICATION_MODE_REQUEST_FRAME */:
                    this._applicationMode = 4 /* APPLICATION_MODE_REQUEST_FRAME */;
                    if (DEBUG_APPLICATION_MODE) {
                        this._log.debug("Starting APPLICATION_MODE_REQUEST_FRAME");
                    }
                    break;
                case "" + 5 /* APPLICATION_MODE_SHOW_FILE */:
                    if (DEBUG_APPLICATION_MODE) {
                        this._log.debug("Starting APPLICATION_MODE_SHOW_FILE");
                    }
                    this._applicationMode = 5 /* APPLICATION_MODE_SHOW_FILE */;
                    this._showData = "";
                    this._showPayloadSize = params[2];
                    break;
                default:
                    this._log.warn("Unrecognized application escape parameters.");
                    break;
            }
        }
    }
    /**
     * Handle incoming data while in application mode.
     *
     * @param {string} data The new data.
     */
    _handleApplicationModeData(data) {
        if (DEBUG_APPLICATION_MODE) {
            this._log.debug("html-mode data!", data);
        }
        switch (this._applicationMode) {
            case 2 /* APPLICATION_MODE_OUTPUT_BRACKET_START */:
            case 3 /* APPLICATION_MODE_OUTPUT_BRACKET_END */:
            case 4 /* APPLICATION_MODE_REQUEST_FRAME */:
                this._htmlData = this._htmlData + data;
                break;
            case 5 /* APPLICATION_MODE_SHOW_FILE */:
                this._showData = this._showData + data;
                break;
            default:
                break;
        }
    }
    /**
     * Handle the exit from application mode.
     */
    _handleApplicationModeEnd() {
        let el;
        let startdivs;
        switch (this._applicationMode) {
            case 1 /* APPLICATION_MODE_HTML */:
                // el = this._getWindow().document.createElement("div");
                // el.innerHTML = this._htmlData;
                // this._appendElementToScrollArea(el);
                break;
            case 2 /* APPLICATION_MODE_OUTPUT_BRACKET_START */:
                this._handleApplicationModeBracketStart();
                break;
            case 3 /* APPLICATION_MODE_OUTPUT_BRACKET_END */:
                this._handleApplicationModeBracketEnd();
                break;
            case 4 /* APPLICATION_MODE_REQUEST_FRAME */:
                this.handleRequestFrame(this._htmlData);
                break;
            case 5 /* APPLICATION_MODE_SHOW_FILE */:
                this._handleShowFile(this._showPayloadSize, this._showData);
                this._showPayloadSize = "";
                this._showData = "";
                break;
            default:
                break;
        }
        this._applicationMode = 0 /* APPLICATION_MODE_NONE */;
        if (DEBUG_APPLICATION_MODE) {
            this._log.debug("html-mode end!", this._htmlData);
        }
        this._htmlData = null;
    }
    _handleApplicationModeBracketStart() {
        for (const kidInfo of this._childElementList) {
            const element = kidInfo.element;
            if ((EmbeddedViewer_1.EmbeddedViewer.is(element) && element.getReturnCode() == null) || CommandPlaceholder_1.CommandPlaceHolder.is(element)) {
                return; // Don't open a new frame.
            }
        }
        // Fetch the command line.
        let cleancommand = this._htmlData;
        if (this._bracketStyle === "bash") {
            // Bash includes the history number. Remove it.
            const trimmed = this._htmlData.trim();
            cleancommand = trimmed.slice(trimmed.indexOf(" ")).trim();
        }
        if (!this._isNoFrameCommand(cleancommand)) {
            // Create and set up a new command-frame.
            const el = this._createEmbeddedViewerElement(cleancommand);
            this._appendScrollableElement(el);
        }
        else {
            this._moveCursorToFreshLine();
            this._emulator.moveRowsAboveCursorToScrollback();
            this._emulator.flushRenderQueue();
            this._lastCommandTerminalLine = this._terminalViewer.bookmarkLine(this._terminalViewer.lineCount() - 1);
            this._lastCommandLine = cleancommand;
            this._lastCommandTerminalViewer = this._terminalViewer;
        }
        const scrollContainer = DomUtils.getShadowId(this, ID_SCROLL_CONTAINER);
        this._virtualScrollArea.updateContainerHeight(scrollContainer.getBoundingClientRect().height);
    }
    _moveCursorToFreshLine() {
        const dims = this._emulator.getDimensions();
        if (dims.cursorX !== 0 && this._emulator.getLineText(dims.cursorY).trim() !== '') {
            this._emulator.newLine();
        }
    }
    deleteEmbeddedViewer(viewer) {
        this._removeScrollable(viewer);
    }
    _getLastEmbeddedViewer() {
        const kids = this._childElementList;
        const len = this._childElementList.length;
        for (let i = len - 1; i >= 0; i--) {
            const kid = kids[i].element;
            if (EmbeddedViewer_1.EmbeddedViewer.is(kid)) {
                return kid;
            }
        }
        return null;
    }
    _deleteLastEmbeddedViewer() {
        const kid = this._getLastEmbeddedViewer();
        if (kid === null) {
            return;
        }
        this.deleteEmbeddedViewer(kid);
        this.focus();
    }
    _createEmbeddedViewerElement(title) {
        // Create and set up a new command-frame.
        const el = this._getWindow().document.createElement(EmbeddedViewer_1.EmbeddedViewer.TAG_NAME);
        KeyBindingManager_1.injectKeyBindingManager(el, this._keyBindingManager);
        Config_1.injectConfigManager(el, this._configManager);
        el.setAwesomeIcon('cog');
        el.addEventListener(EmbeddedViewer_1.EmbeddedViewer.EVENT_CLOSE_REQUEST, () => {
            this.deleteEmbeddedViewer(el);
            this.focus();
        });
        el.addEventListener(GeneralEvents.EVENT_SET_MODE, (ev) => {
            const detail = ev.detail;
            if (detail.mode !== this._mode) {
                switch (this._mode) {
                    case Mode.DEFAULT:
                        this._enterCursorMode();
                        break;
                    case Mode.CURSOR:
                        this._exitCursorMode();
                        break;
                }
            }
        });
        el.addEventListener(EmbeddedViewer_1.EmbeddedViewer.EVENT_FRAME_POP_OUT, (ev) => {
            this._embeddedViewerPopOutEvent(ev.srcElement);
            ev.stopPropagation();
        });
        // el.addEventListener('copy-clipboard-request', (function(ev: CustomEvent) {
        //   var clipboard = gui.Clipboard.get();
        //   clipboard.set(ev.detail, 'text');
        // }).bind(this));
        // FIXME
        el.setVisualState(DomUtils.getShadowRoot(this).activeElement !== null
            ? VisualState.FOCUSED
            : VisualState.UNFOCUSED);
        el.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_FRAME_TITLE, title);
        el.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_TAG, "" + this._getNextTag());
        return el;
    }
    _handleApplicationModeBracketEnd() {
        this._enforceScrollbackLengthAfter(() => {
            this._closeLastEmbeddedViewer(this._htmlData);
        });
    }
    _closeLastEmbeddedViewer(returnCode) {
        // Find the terminal viewer which has no return code set on it.
        let startElement = null;
        for (let i = this._childElementList.length - 1; i >= 0; i--) {
            const el = this._childElementList[i].element;
            if (el.tagName === EmbeddedViewer_1.EmbeddedViewer.TAG_NAME && el.getAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_RETURN_CODE) == null) {
                startElement = el;
                break;
            }
        }
        if (startElement != null) {
            // Finish framing an already existing Embedded viewer bar.
            const embeddedViewerElement = startElement;
            const activeTerminalViewer = this._terminalViewer;
            this._disconnectActiveTerminalViewer();
            activeTerminalViewer.setReturnCode(returnCode);
            activeTerminalViewer.setCommandLine(embeddedViewerElement.getAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_FRAME_TITLE));
            activeTerminalViewer.setUseVPad(false);
            // Hang the terminal viewer under the Embedded viewer.
            embeddedViewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_RETURN_CODE, returnCode);
            embeddedViewerElement.setAwesomeIcon(returnCode === '0' ? 'check' : 'times');
            embeddedViewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_TOOL_TIP, "Return code: " + returnCode);
            embeddedViewerElement.className = "extraterm_output";
            // Some focus management to make sure that activeTerminalViewer still keeps
            // the focus after we remove it from the DOM and place it else where.
            const restoreFocus = DomUtils.getShadowRoot(this).activeElement === activeTerminalViewer;
            embeddedViewerElement.setViewerElement(activeTerminalViewer);
            activeTerminalViewer.setEditable(true);
            this._removeScrollable(activeTerminalViewer);
            this._virtualScrollArea.updateScrollableSize(embeddedViewerElement);
            this._appendNewTerminalViewer();
            if (restoreFocus) {
                this._terminalViewer.focus();
            }
        }
        else {
            if (this._lastCommandTerminalViewer === null) {
                // Nothing to frame.
                return;
            }
            if (returnCode === "0") {
                // No need to frame anything.
                this._lastCommandLine = null;
                this._lastCommandTerminalViewer = null;
                return;
            }
            // Insert a frame where there was none because this command returned an error code.
            // Close off the current terminal viewer.
            this._disconnectActiveTerminalViewer();
            // Extract the output of the failed command.
            const moveText = this._lastCommandTerminalViewer.getDecoratedLines(this._lastCommandTerminalLine);
            this._lastCommandTerminalViewer.deleteLines(this._lastCommandTerminalLine);
            this._lastCommandTerminalViewer = null;
            // Append our new embedded viewer.
            const newViewerElement = this._createEmbeddedViewerElement(this._lastCommandLine);
            // Hang the terminal viewer under the Embedded viewer.
            newViewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_RETURN_CODE, returnCode);
            newViewerElement.setAwesomeIcon('times');
            newViewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_TOOL_TIP, "Return code: " + returnCode);
            newViewerElement.className = "extraterm_output";
            this._appendScrollable(newViewerElement);
            // Create a terminal viewer to display the output of the last command.
            const outputTerminalViewer = document.createElement(TerminalViewer_1.TerminalViewer.TAG_NAME);
            KeyBindingManager_1.injectKeyBindingManager(outputTerminalViewer, this._keyBindingManager);
            Config_1.injectConfigManager(outputTerminalViewer, this._configManager);
            newViewerElement.setViewerElement(outputTerminalViewer);
            outputTerminalViewer.setVisualState(DomUtils.getShadowRoot(this).activeElement !== null
                ? VisualState.FOCUSED
                : VisualState.UNFOCUSED);
            outputTerminalViewer.setReturnCode(returnCode);
            outputTerminalViewer.setCommandLine(this._lastCommandLine);
            outputTerminalViewer.setUseVPad(false);
            if (moveText !== null) {
                outputTerminalViewer.setDecoratedLines(moveText.text, moveText.decorations);
            }
            outputTerminalViewer.setEditable(true);
            this._appendNewTerminalViewer();
            this._refocus();
            const activeTerminalViewer = this._terminalViewer;
            this._virtualScrollArea.updateScrollableSize(activeTerminalViewer);
        }
    }
    // ********************************************************************
    //
    //   #     #                 
    //   ##   ## #  ####   ####  
    //   # # # # # #      #    # 
    //   #  #  # #  ####  #      
    //   #     # #      # #      
    //   #     # # #    # #    # 
    //   #     # #  ####   ####  
    //
    // ********************************************************************
    /**
     * Copy the selection to the clipboard.
     */
    copyToClipboard() {
        let text = null;
        for (let i = 0; i < this._childElementList.length; i++) {
            const node = this._childElementList[i].element;
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                text = node.getSelectionText();
                if (text !== null) {
                    WebIpc.clipboardWrite(text);
                    break;
                }
            }
        }
    }
    canPaste() {
        return true;
    }
    pasteText(text) {
        if (this._mode === Mode.CURSOR) {
            const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
            for (const node of scrollerArea.childNodes) {
                if (ViewerElement_1.ViewerElement.isViewerElement(node) && node.hasFocus()) {
                    if (SupportsClipboardPaste.isSupportsClipboardPaste(node) && node.canPaste()) {
                        node.pasteText(text);
                    }
                    break;
                }
            }
        }
        else {
            this.send(text); // Send it to the PTY.
        }
    }
    /**
     * Paste text from the clipboard.
     *
     * This method is async and returns before the paste is done.
     */
    _pasteFromClipboard() {
        WebIpc.clipboardReadRequest();
    }
    _embeddedViewerPopOutEvent(viewerElement) {
        const event = new CustomEvent(EtTerminal.EVENT_EMBEDDED_VIEWER_POP_OUT, { detail: { terminal: this, embeddedViewer: viewerElement } });
        this.dispatchEvent(event);
    }
    handleRequestFrame(frameId) {
        let data = "";
        if (this._frameFinder !== null) {
            data = this._frameFinder(frameId);
        }
        else {
            data = this.getFrameContents(frameId);
        }
        // FIXME this should be binary safe and send metadata like mimetype and charset.
        const lines = data == null ? [] : data.split("\n");
        let encodedData = "";
        lines.forEach((line) => {
            const utf8line = utf8.encode(line);
            encodedData = window.btoa(utf8line + "\n");
            this._sendDataToPtyEvent("#" + encodedData + "\n");
        });
        this._sendDataToPtyEvent("#;0\n"); // Terminating char
    }
    _handleShowFile(metadataSizeStr, encodedData) {
        const metadataSize = parseInt(metadataSizeStr, 10);
        if (metadataSize > encodedData.length) {
            this._log.warn("Received corrupt data for a 'show' control sequence.");
            return;
        }
        const buffer = Buffer.from(encodedData.slice(metadataSize), 'base64');
        const metadata = JSON.parse(encodedData.substr(0, metadataSize));
        const filename = metadata.filename;
        let mimeType = metadata.mimeType || null;
        let charset = metadata.charset || null;
        if (mimeType === null) {
            // Try to determine a mimetype by inspecting the file name first.
            const detectionResult = MmeTypeDetector.detect(filename, buffer);
            if (detectionResult !== null) {
                mimeType = detectionResult.mimeType;
                if (charset === null) {
                    charset = detectionResult.charset;
                }
            }
        }
        if (mimeType !== null) {
            this._appendMimeViewer(mimeType, filename, charset, buffer);
        }
    }
    _appendMimeViewer(mimeType, filename, charset, data) {
        const mimeViewerElement = this._createMimeViewer(mimeType, charset, data);
        if (mimeViewerElement !== null) {
            this._closeLastEmbeddedViewer("0");
            const viewerElement = this._createEmbeddedViewerElement("viewer");
            viewerElement.setViewerElement(mimeViewerElement);
            viewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_FRAME_TITLE, filename);
            viewerElement.setAwesomeIcon(mimeViewerElement.getAwesomeIcon());
            viewerElement.setAttribute(EmbeddedViewer_1.EmbeddedViewer.ATTR_RETURN_CODE, "0"); // FIXME
            this._appendScrollableElement(viewerElement);
            this._enforceScrollbackLength(this._scrollbackSize);
        }
    }
    _createMimeViewer(mimeType, charset, data) {
        const candidates = viewerClasses.filter((viewerClass) => viewerClass.supportsMimeType(mimeType));
        if (candidates.length === 0) {
            this._log.debug("Unknown mime type: " + mimeType);
            return null;
        }
        const dataViewer = this._getWindow().document.createElement(candidates[0].TAG_NAME);
        KeyBindingManager_1.injectKeyBindingManager(dataViewer, this._keyBindingManager);
        Config_1.injectConfigManager(dataViewer, this._configManager);
        if (data !== null) {
            dataViewer.setBytes(data, charset !== null ? mimeType + ";" + charset : mimeType);
        }
        dataViewer.setEditable(true);
        return dataViewer;
    }
    _getNextTag() {
        let tag = this._nextTag;
        if (tag === null) {
            // Fetching new tags from the main process is async. If we get here it means
            // that we were waiting for a new tag to arrive. Just fetch one sync.
            tag = WebIpc.requestNewTagSync();
        }
        this._fetchNextTag();
        return tag;
    }
    _fetchNextTag() {
        WebIpc.requestNewTag().then((msg) => {
            this._nextTag = msg.tag;
        });
    }
}
/**
 * The HTML tag name of this element.
 */
EtTerminal.TAG_NAME = "ET-TERMINAL";
EtTerminal.EVENT_USER_INPUT = "user-input";
EtTerminal.EVENT_TERMINAL_RESIZE = "terminal-resize";
EtTerminal.EVENT_TERMINAL_BUFFER_SIZE = "temrinal-buffer-size";
EtTerminal.EVENT_TITLE = "title";
EtTerminal.EVENT_EMBEDDED_VIEWER_POP_OUT = "viewer-pop-out";
exports.EtTerminal = EtTerminal;
//# sourceMappingURL=Terminal.js.map