"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeTypes = require("./Theme");
/**
 * Module for handling the distribution of newly computed CSS themes to all
 * elements which need them. This is the render process based counter-part
 * to the ThemeManager. It receives the updated CSS from ThemeManager in the
 * main process.
 */
// Our simple registry of instances to update.
const themeableRegistry = new Set();
// Last known CssFile map.
let currentCssMap = new Map();
ThemeTypes.cssFileEnumItems.forEach((cssFile) => {
    currentCssMap.set(cssFile, "");
});
/**
 * Register a `Themeable` instance for updates.
 *
 * @param themeable the instance which requires theme CSS updates
 */
function registerThemeable(themeable) {
    themeableRegistry.add(themeable);
}
exports.registerThemeable = registerThemeable;
/**
 * Unregister a `Themeable` instance.
 *
 * This method should be called to unregister a themeable instance before
 * it is destroyed. Failing to unregister may result in a memory leak.
 *
 * @param themeable the instance to unregister
 */
function unregisterThemeable(themeable) {
    themeableRegistry.delete(themeable);
}
exports.unregisterThemeable = unregisterThemeable;
/**
 * Updates the CSS used by all Themeable instances with newly computed CSS.
 *
 * @param cssMap the map of CssFiles to their matching CSS texts
 */
function updateCss(cssMap) {
    const newCssMap = new Map(currentCssMap);
    cssMap.forEach((value, key) => newCssMap.set(key, value));
    currentCssMap = newCssMap;
    themeableRegistry.forEach((themeable) => themeable.setThemeCssMap(newCssMap));
}
exports.updateCss = updateCss;
/**
 * Gets the map of CssFiles to their CSS texts.
 *
 * @returns the map
 */
function cssMap() {
    return currentCssMap;
}
exports.cssMap = cssMap;
//# sourceMappingURL=ThemeConsumer.js.map