"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ViewerElement_1 = require("./ViewerElement");
const ViewerElementTypes = require("./ViewerElementTypes");
const ResizeRefreshElementBase = require("./ResizeRefreshElementBase");
const EmbeddedViewer_1 = require("./EmbeddedViewer");
const Logger_1 = require("./Logger");
const DomUtils = require("./DomUtils");
const ScrollBar_1 = require("./gui/ScrollBar");
const ResizeCanary_1 = require("./ResizeCanary");
const ThemeableElementBase_1 = require("./ThemeableElementBase");
const ThemeTypes = require("./Theme");
const SupportsClipboardPaste = require("./SupportsClipboardPaste");
const CommandPaletteRequestTypes_1 = require("./CommandPaletteRequestTypes");
const WebIpc = require("./WebIpc");
const VirtualScrollArea = require("./VirtualScrollArea");
const VisualState = ViewerElementTypes.VisualState;
const DEBUG = true;
let registered = false;
const ID = "EtTabViewerTemplate";
const ID_SCROLL_AREA = "ID_SCROLL_AREA";
const ID_SCROLLBAR = "ID_SCROLLBAR";
const ID_CONTAINER = "ID_CONTAINER";
const ID_CSS_VARS = "ID_CSS_VARS";
const KEYBINDINGS_VIEWER_TAB = "viewer-tab";
const PALETTE_GROUP = "viewertab";
const COMMAND_COPY_TO_CLIPBOARD = "copyToClipboard";
const COMMAND_PASTE_FROM_CLIPBOARD = "pasteFromClipboard";
const COMMAND_FONT_SIZE_INCREASE = "increaseFontSize";
const COMMAND_FONT_SIZE_DECREASE = "decreaseFontSize";
const COMMAND_FONT_SIZE_RESET = "resetFontSize";
const MINIMUM_FONT_SIZE = -3;
const MAXIMUM_FONT_SIZE = 4;
const SCROLL_STEP = 1;
/**
 * A viewer tab which can contain any ViewerElement.
 */
class EtViewerTab extends ViewerElement_1.ViewerElement {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    /**
     * Initialize the EtViewerTab class and resources.
     *
     * When EtViewerTab is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            ScrollBar_1.ScrollBar.init();
            EmbeddedViewer_1.EmbeddedViewer.init();
            ResizeCanary_1.ResizeCanary.init();
            window.customElements.define(EtViewerTab.TAG_NAME.toLowerCase(), EtViewerTab);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger(EtViewerTab.TAG_NAME, this);
        this._virtualScrollArea = null;
        this._elementAttached = false;
        this._needsCompleteRefresh = true;
        this._blinkingCursor = false;
        this._title = "New Tab";
        this._tag = null;
        this._mainStyleLoaded = false;
        this._themeStyleLoaded = false;
        this._resizePollHandle = null;
        this._terminalSize = null;
        this._scrollYOffset = 0;
        this._virtualHeight = 0;
        this._fontSizeAdjustment = 0;
        this._armResizeCanary = false;
        this._scheduleLaterHandle = null;
        this._scheduledResize = false;
    }
    //-----------------------------------------------------------------------
    //
    //   ######                                
    //   #     # #    # #####  #      #  ####  
    //   #     # #    # #    # #      # #    # 
    //   ######  #    # #####  #      # #      
    //   #       #    # #    # #      # #      
    //   #       #    # #    # #      # #    # 
    //   #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    /**
     * Get this terminal's title.
     *
     * This is the window title of the terminal, don't confuse it with more
     * general HTML title of the element.
     */
    getTerminalTitle() {
        return this._title;
    }
    getTitle() {
        return this._title;
    }
    setTitle(newTitle) {
        this._title = newTitle;
    }
    getTag() {
        return this._tag;
    }
    setTag(tag) {
        this._tag = tag;
    }
    /**
     * Destroy the ViewerTab
     */
    destroy() {
        if (this._resizePollHandle !== null) {
            this._resizePollHandle.cancel();
            this._resizePollHandle = null;
        }
        // this._getWindow().removeEventListener('resize', this._scheduleResize.bind(this));
    }
    /**
     * Focus on this ViewerTab.
     */
    focus() {
        const element = this.getViewerElement();
        if (element !== null) {
            element.focus();
        }
    }
    /**
     * Returns true if this terminal has the input focus.
     *
     * @return true if the terminal has the focus.
     */
    hasFocus() {
        const shadowRoot = DomUtils.getShadowRoot(this);
        if (shadowRoot === null) {
            return false;
        }
        return shadowRoot.activeElement !== null;
    }
    setViewerElement(element) {
        if (this.childNodes.length !== 0) {
            this.innerHTML = "";
        }
        if (element !== null) {
            // element.visualState = ViewerElementTypes. this._visualState; FIXME
            element.setMode(ViewerElementTypes.Mode.CURSOR);
            this._appendScrollableElement(element);
        }
    }
    getViewerElement() {
        return this._getViewerElement();
    }
    getAwesomeIcon() {
        const viewerElement = this.getViewerElement();
        return viewerElement === null ? "desktop" : viewerElement.getAwesomeIcon();
    }
    getFrameContents(frameId) {
        const viewerElement = this.getViewerElement();
        if (viewerElement === null) {
            return null;
        }
        if (this._tag === frameId) {
            return viewerElement.getText();
        }
        else {
            return null;
        }
    }
    getMode() {
        return ViewerElementTypes.Mode.CURSOR;
    }
    setKeyBindingManager(keyBindingManager) {
        this._keyBindingManager = keyBindingManager;
    }
    getFontAdjust() {
        return this._fontSizeAdjustment;
    }
    setFontAdjust(delta) {
        this._adjustFontSize(delta);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.VIEWER_TAB];
    }
    /**
     * Custom Element 'connected' life cycle hook.
     */
    connectedCallback() {
        super.connectedCallback();
        if (this._elementAttached) {
            return;
        }
        this._elementAttached = true;
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this._createClone();
        shadow.appendChild(clone);
        this._virtualScrollArea = new VirtualScrollArea.VirtualScrollArea();
        this.addEventListener('focus', this._handleFocus.bind(this));
        this.addEventListener('blur', this._handleBlur.bind(this));
        const scrollbar = DomUtils.getShadowId(this, ID_SCROLLBAR);
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        const scrollContainer = DomUtils.getShadowId(this, ID_CONTAINER);
        DomUtils.preventScroll(scrollContainer);
        this._virtualScrollArea.setScrollFunction((offset) => {
            scrollerArea.scrollTop = offset;
        });
        this._virtualScrollArea.setScrollbar(scrollbar);
        scrollerArea.addEventListener('wheel', this._handleMouseWheel.bind(this), true);
        scrollerArea.addEventListener('mousedown', (ev) => {
            if (ev.target === scrollerArea) {
                ev.preventDefault();
                ev.stopPropagation();
                this.focus();
                if (ev.buttons & 2) {
                    const viewerElement = this._getViewerElement();
                    if (viewerElement !== null && CommandPaletteRequestTypes_1.isCommandable(viewerElement)) {
                        viewerElement.executeCommand(CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE);
                    }
                }
            }
        });
        scrollbar.addEventListener('scroll', (ev) => {
            this._virtualScrollArea.scrollTo(scrollbar.getPosition());
        });
        scrollerArea.addEventListener('mousedown', this._handleMouseDown.bind(this), true);
        scrollerArea.addEventListener('keydown', this._handleKeyDownCapture.bind(this), true);
        scrollerArea.addEventListener(VirtualScrollArea.EVENT_RESIZE, this._handleVirtualScrollableResize.bind(this));
        scrollerArea.addEventListener(ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE, this._handleTerminalViewerCursor.bind(this));
        scrollerArea.addEventListener(ViewerElement_1.ViewerElement.EVENT_BEFORE_SELECTION_CHANGE, this._handleBeforeSelectionChange.bind(this));
        // A Resize Canary for tracking when terminal fonts are effectively changed in the DOM.
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        const resizeCanary = document.createElement(ResizeCanary_1.ResizeCanary.TAG_NAME);
        resizeCanary.setCss(`
        font-family: var(--terminal-font);
        font-size: var(--terminal-font-size);
    `);
        containerDiv.appendChild(resizeCanary);
        resizeCanary.addEventListener('resize', () => {
            if (this._armResizeCanary) {
                this._armResizeCanary = false;
                this.refresh(ResizeRefreshElementBase.RefreshLevel.COMPLETE);
            }
        });
        this.updateThemeCss();
        DomUtils.doLater(this._processResize.bind(this));
    }
    /**
     * Custom Element 'disconnected' life cycle hook.
     */
    disconnectedCallback() {
        super.disconnectedCallback();
        this._needsCompleteRefresh = true;
    }
    refresh(requestedLevel) {
        let level = requestedLevel;
        if (this._needsCompleteRefresh) {
            level = ResizeRefreshElementBase.RefreshLevel.COMPLETE;
            this._needsCompleteRefresh = false;
        }
        const viewerElement = this._getViewerElement();
        if (viewerElement != null) {
            // --- DOM read ---
            const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
            const scrollbar = DomUtils.getShadowId(this, ID_SCROLLBAR);
            ResizeRefreshElementBase.ResizeRefreshElementBase.refreshChildNodes(scrollerArea, level);
            scrollbar.refresh(level);
            // --- DOM write ---
            const scrollContainer = DomUtils.getShadowId(this, ID_CONTAINER);
            this._virtualScrollArea.updateContainerHeight(scrollContainer.getBoundingClientRect().height);
            viewerElement.refresh(level);
            this._virtualScrollArea.updateScrollableSizes([viewerElement]);
            this._virtualScrollArea.reapplyState();
        }
    }
    //-----------------------------------------------------------------------
    //
    //   ######                                      
    //   #     # #####  # #    #   ##   ##### ###### 
    //   #     # #    # # #    #  #  #    #   #      
    //   ######  #    # # #    # #    #   #   #####  
    //   #       #####  # #    # ######   #   #      
    //   #       #   #  #  #  #  #    #   #   #      
    //   #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    _createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `
        <style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
        <style id="${ID_CSS_VARS}">${this._getCssVarsRules()}</style>
        <div id='${ID_CONTAINER}'>
          <div id='${ID_SCROLL_AREA}'></div>
          <${ScrollBar_1.ScrollBar.TAG_NAME} id='${ID_SCROLLBAR}'></${ScrollBar_1.ScrollBar.TAG_NAME}>
        </div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _getCssVarsRules() {
        return `
    #${ID_CONTAINER} {
        ${this._getCssFontSizeRule(this._fontSizeAdjustment)}
    }
    `;
    }
    _getCssFontSizeRule(adjustment) {
        const scale = [0.6, 0.75, 0.89, 1, 1.2, 1.5, 2, 3][adjustment - MINIMUM_FONT_SIZE];
        return `--terminal-font-size: calc(var(--default-terminal-font-size) * ${scale});`;
    }
    /**
     * Get the window which this terminal is on.
     *
     * @returns {Window} The window object.
     */
    _getWindow() {
        return this.ownerDocument.defaultView;
    }
    _getDocument() {
        return this.ownerDocument;
    }
    _handleFocus(event) {
        // Forcefully set the visual state of each thing in the terminal to appear focused.
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        DomUtils.nodeListToArray(scrollerArea.childNodes).forEach((node) => {
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                node.setVisualState(VisualState.FOCUSED);
            }
        });
    }
    _handleBlur(event) {
        // Forcefully set the visual state of each thing in the terminal to appear unfocused.
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        DomUtils.nodeListToArray(scrollerArea.childNodes).forEach((node) => {
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                node.setVisualState(VisualState.UNFOCUSED);
            }
        });
    }
    _getViewerElement() {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        if (scrollerArea.firstElementChild !== null && scrollerArea.firstElementChild instanceof ViewerElement_1.ViewerElement) {
            return scrollerArea.firstElementChild;
        }
        else {
            return null;
        }
    }
    _appendScrollableElement(el) {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        scrollerArea.appendChild(el);
        this._virtualScrollArea.appendScrollable(el);
    }
    _handleMouseDown(ev) {
        if (ev.buttons === 4) {
            ev.stopPropagation();
            ev.preventDefault();
            this._pasteFromClipboard();
        }
    }
    // ----------------------------------------------------------------------
    //
    //    #####                                                          ##        #####                           
    //   #     #  ####  #####   ####  #      #      # #    #  ####      #  #      #     # # ###### # #    #  ####  
    //   #       #    # #    # #    # #      #      # ##   # #    #      ##       #       #     #  # ##   # #    # 
    //    #####  #      #    # #    # #      #      # # #  # #          ###        #####  #    #   # # #  # #      
    //         # #      #####  #    # #      #      # #  # # #  ###    #   # #          # #   #    # #  # # #  ### 
    //   #     # #    # #   #  #    # #      #      # #   ## #    #    #    #     #     # #  #     # #   ## #    # 
    //    #####   ####  #    #  ####  ###### ###### # #    #  ####      ###  #     #####  # ###### # #    #  ####  
    //
    // ----------------------------------------------------------------------
    _handleMouseWheel(ev) {
        ev.stopPropagation();
        ev.preventDefault();
        const delta = ev.deltaY * SCROLL_STEP;
        this._virtualScrollArea.scrollTo(this._virtualScrollArea.getScrollYOffset() + delta);
    }
    _handleVirtualScrollableResize(ev) {
        this._updateVirtualScrollableSize(ev.target);
        // ^ We know this event only comes from VirtualScrollable elements.
    }
    _updateVirtualScrollableSize(virtualScrollable) {
        this._virtualScrollArea.updateScrollableSize(virtualScrollable);
    }
    /**
     * Handle a resize event from the window.
     */
    _processResize() {
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        this._virtualScrollArea.updateContainerHeight(scrollerArea.getBoundingClientRect().height);
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            this._updateVirtualScrollableSize(viewerElement);
        }
    }
    _handleTerminalViewerCursor(ev) {
        const node = ev.target;
        if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
            const pos = node.getCursorPosition();
            const nodeTop = this._virtualScrollArea.getScrollableTop(node);
            const top = pos.top + nodeTop;
            const bottom = pos.bottom;
            this._virtualScrollArea.scrollIntoView(top, bottom);
        }
    }
    _adjustFontSize(delta) {
        const newAdjustment = Math.min(Math.max(this._fontSizeAdjustment + delta, MINIMUM_FONT_SIZE), MAXIMUM_FONT_SIZE);
        if (newAdjustment !== this._fontSizeAdjustment) {
            this._fontSizeAdjustment = newAdjustment;
            const styleElement = DomUtils.getShadowId(this, ID_CSS_VARS);
            if (styleElement != null) {
                styleElement.sheet.cssRules[0].style.cssText = this._getCssFontSizeRule(newAdjustment); // Type stubs are missing cssRules.
                this._armResizeCanary = true;
                // Don't refresh. Let the Resize Canary detect the real change in the DOM when it arrives.
            }
        }
    }
    _resetFontSize() {
        this._adjustFontSize(-this._fontSizeAdjustment);
    }
    // ----------------------------------------------------------------------
    //
    //   #    #                                                 
    //   #   #  ###### #   # #####   ####    ##   #####  #####  
    //   #  #   #       # #  #    # #    #  #  #  #    # #    # 
    //   ###    #####    #   #####  #    # #    # #    # #    # 
    //   #  #   #        #   #    # #    # ###### #####  #    # 
    //   #   #  #        #   #    # #    # #    # #   #  #    # 
    //   #    # ######   #   #####   ####  #    # #    # #####  
    //                                                        
    // ----------------------------------------------------------------------
    _handleKeyDownCapture(ev) {
        if (this._keyBindingManager === null || this._keyBindingManager.getKeyBindingContexts() === null) {
            return;
        }
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_VIEWER_TAB);
        const command = keyBindings.mapEventToCommand(ev);
        if (this._executeCommand(command)) {
            ev.stopPropagation();
            ev.preventDefault();
        }
    }
    getCommandPaletteEntries(commandableStack) {
        const commandList = [];
        commandList.push({ id: COMMAND_FONT_SIZE_INCREASE, group: PALETTE_GROUP, label: "Increase Font Size", commandExecutor: this });
        commandList.push({ id: COMMAND_FONT_SIZE_DECREASE, group: PALETTE_GROUP, label: "Decrease Font Size", commandExecutor: this });
        commandList.push({ id: COMMAND_FONT_SIZE_RESET, group: PALETTE_GROUP, label: "Reset Font Size", commandExecutor: this });
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_VIEWER_TAB);
        if (keyBindings !== null) {
            commandList.forEach((commandEntry) => {
                const shortcut = keyBindings.mapCommandToKeyBinding(commandEntry.id);
                commandEntry.shortcut = shortcut === null ? "" : shortcut;
            });
        }
        return commandList;
    }
    executeCommand(commandId) {
        this._executeCommand(commandId);
    }
    _executeCommand(command) {
        switch (command) {
            case COMMAND_COPY_TO_CLIPBOARD:
                this.copyToClipboard();
                break;
            case COMMAND_PASTE_FROM_CLIPBOARD:
                this._pasteFromClipboard();
                break;
            case COMMAND_FONT_SIZE_INCREASE:
                this._adjustFontSize(1);
                break;
            case COMMAND_FONT_SIZE_DECREASE:
                this._adjustFontSize(-1);
                break;
            case COMMAND_FONT_SIZE_RESET:
                this._resetFontSize();
                break;
            default:
                return false;
        }
        return true;
    }
    // ********************************************************************
    //
    //   #     #                 
    //   ##   ## #  ####   ####  
    //   # # # # # #      #    # 
    //   #  #  # #  ####  #      
    //   #     # #      # #      
    //   #     # # #    # #    # 
    //   #     # #  ####   ####  
    //
    // ********************************************************************
    _handleBeforeSelectionChange(ev) {
        if (ev.detail.originMouse) {
            DomUtils.doLater(() => { this.copyToClipboard(); });
        }
    }
    canPaste() {
        const viewerElement = this.getViewerElement();
        if (viewerElement === null) {
            return false;
        }
        if (SupportsClipboardPaste.isSupportsClipboardPaste(viewerElement)) {
            return viewerElement.canPaste();
        }
        return false;
    }
    pasteText(text) {
        const viewerElement = this.getViewerElement();
        if (viewerElement != null && SupportsClipboardPaste.isSupportsClipboardPaste(viewerElement)) {
            viewerElement.pasteText(text);
        }
    }
    /**
     * Copy the selection to the clipboard.
     */
    copyToClipboard() {
        let text = null;
        const scrollerArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        const kids = DomUtils.nodeListToArray(scrollerArea.childNodes);
        for (let i = 0; i < kids.length; i++) {
            const node = kids[i];
            if (ViewerElement_1.ViewerElement.isViewerElement(node)) {
                text = node.getSelectionText();
                if (text !== null) {
                    WebIpc.clipboardWrite(text);
                    break;
                }
            }
        }
    }
    /**
     * Paste text from the clipboard.
     *
     * This method is async and returns before the paste is done.
     */
    _pasteFromClipboard() {
        WebIpc.clipboardReadRequest();
    }
    /**
     * Find a command frame by ID.
     */
    _findFrame(frameId) {
        if (/[^0-9]/.test(frameId)) {
            return null;
        }
        const scrollArea = DomUtils.getShadowId(this, ID_SCROLL_AREA);
        const matches = scrollArea.querySelectorAll(EmbeddedViewer_1.EmbeddedViewer.TAG_NAME + "[tag='" + frameId + "']");
        return matches.length === 0 ? null : matches[0];
    }
}
/**
 * The HTML tag name of this element.
 */
EtViewerTab.TAG_NAME = "ET-VIEWER-TAB";
exports.EtViewerTab = EtViewerTab;
//# sourceMappingURL=ViewerTab.js.map