"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const Electron = require("electron");
const ipc = Electron.ipcRenderer;
const Messages = require("./WindowMessages");
const Logger_1 = require("./Logger");
const _log = Logger_1.getLogger("WebIPC");
const DEBUG = false;
const DEBUG_INCOMING = false;
const DEBUG_OUTGOING = false;
/**
 * Start IPC.
 */
function start() {
    ipc.on(Messages.CHANNEL_NAME, handleAsyncIpc);
}
exports.start = start;
// This queue is used to route incoming IPC messages to expectant promise objects.
let promiseQueue = [];
const defaultHandlers = [];
/**
 * Register a default handler for a message type
 *
 * @param type    the message type this handler should be used for.
 * @param handler the handler itself.
 */
function registerDefaultHandler(type, handler) {
    defaultHandlers.push({ messageType: type, handler: handler });
}
exports.registerDefaultHandler = registerDefaultHandler;
function handleAsyncIpc(senderInfo, detail) {
    const msg = detail;
    if (DEBUG_INCOMING) {
        _log.debug("incoming: ", msg);
    }
    const matchingPromises = promiseQueue.filter(p => p.messageType === msg.type);
    const nonMatchingPromises = promiseQueue.filter(p => p.messageType !== msg.type);
    promiseQueue = nonMatchingPromises;
    matchingPromises.forEach(tup => {
        tup.promiseResolve(msg);
    });
    if (matchingPromises.length === 0) {
        // Fall back on the default handlers.
        defaultHandlers.filter(tup => tup.messageType === msg.type).forEach(tup => {
            tup.handler(msg);
        });
    }
}
function request(msg, replyType) {
    if (DEBUG_OUTGOING) {
        _log.debug("request: ", msg);
    }
    ipc.send(Messages.CHANNEL_NAME, msg);
    return new Promise((resolve, cancel) => {
        promiseQueue.push({ promiseResolve: resolve, messageType: replyType });
    });
}
function requestConfig() {
    const msg = { type: 0 /* CONFIG_REQUEST */ };
    return request(msg, 1 /* CONFIG */);
}
exports.requestConfig = requestConfig;
function requestThemeList() {
    const msg = { type: 4 /* THEME_LIST_REQUEST */ };
    return request(msg, 5 /* THEME_LIST */);
}
exports.requestThemeList = requestThemeList;
function requestThemeContents(themeIdList, cssFileList) {
    if (DEBUG) {
        _log.debug("requestThemeContents(): ", themeIdList);
    }
    const msg = { type: 6 /* THEME_CONTENTS_REQUEST */,
        themeIdList: themeIdList,
        cssFileList: cssFileList };
    return request(msg, 7 /* THEME_CONTENTS */);
}
exports.requestThemeContents = requestThemeContents;
function requestPtyCreate(command, args, columns, rows, env) {
    if (args === undefined) {
        args = [];
    }
    const msg = {
        type: 8 /* PTY_CREATE */,
        command: command,
        args: args,
        columns: columns,
        rows: rows,
        env: env
    };
    return request(msg, 9 /* PTY_CREATED */);
}
exports.requestPtyCreate = requestPtyCreate;
function ptyInput(id, data) {
    const msg = { type: 12 /* PTY_INPUT */, id: id, data: data };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.ptyInput = ptyInput;
function ptyOutputBufferSize(id, size) {
    const msg = { type: 25 /* PTY_OUTPUT_BUFFER_SIZE */, id, size };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.ptyOutputBufferSize = ptyOutputBufferSize;
function ptyResize(id, columns, rows) {
    const msg = { type: 10 /* PTY_RESIZE */, id: id, columns: columns, rows: rows };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.ptyResize = ptyResize;
function ptyClose(id) {
    const msg = { type: 14 /* PTY_CLOSE_REQUEST */, id: id };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.ptyClose = ptyClose;
function devToolsRequest(open) {
    const msg = { type: 15 /* DEV_TOOLS_REQUEST */, open: open };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.devToolsRequest = devToolsRequest;
function clipboardWrite(text) {
    const msg = { type: 17 /* CLIPBOARD_WRITE */, text: text };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.clipboardWrite = clipboardWrite;
function clipboardReadRequest() {
    const msg = { type: 18 /* CLIPBOARD_READ_REQUEST */ };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.clipboardReadRequest = clipboardReadRequest;
function windowCloseRequest() {
    const msg = { type: 20 /* WINDOW_CLOSE_REQUEST */ };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.windowCloseRequest = windowCloseRequest;
function sendConfig(config) {
    const msg = { type: 1 /* CONFIG */, config: config };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.sendConfig = sendConfig;
function requestNewTag() {
    const msg = { type: 23 /* NEW_TAG_REQUEST */, async: true };
    return request(msg, 24 /* NEW_TAG */);
}
exports.requestNewTag = requestNewTag;
function requestNewTagSync() {
    const msg = { type: 23 /* NEW_TAG_REQUEST */, async: false };
    const event = ipc.sendSync(Messages.CHANNEL_NAME, msg);
    const newTagMessage = event;
    return newTagMessage.tag;
}
exports.requestNewTagSync = requestNewTagSync;
function windowMinimizeRequest() {
    const msg = { type: 21 /* WINDOW_MINIMIZE_REQUEST */ };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.windowMinimizeRequest = windowMinimizeRequest;
function windowMaximizeRequest() {
    const msg = { type: 22 /* WINDOW_MAXIMIZE_REQUEST */ };
    ipc.send(Messages.CHANNEL_NAME, msg);
}
exports.windowMaximizeRequest = windowMaximizeRequest;
//# sourceMappingURL=WebIpc.js.map