"use strict";
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const PopDownDialog_1 = require("./PopDownDialog");
const ID = "EtPopDownNumberDialogTemplate";
const ID_DIALOG = "ID_DIALOG";
const ID_INPUT = "ID_INPUT";
let registered = false;
/**
 * A Pop Down Number Dialog
 */
class PopDownNumberDialog extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties(); // Initialise our properties. The constructor was not called.
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const dialog = DomUtils.getShadowId(this, ID_DIALOG);
        dialog.setTitlePrimary(this._titlePrimary);
        dialog.setTitleSecondary(this._titleSecondary);
        dialog.addEventListener(PopDownDialog_1.PopDownDialog.EVENT_CLOSE_REQUEST, () => {
            dialog.close();
        });
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        textInput.addEventListener('keydown', (ev) => { this.handleKeyDown(ev); });
    }
    /**
     * Initialize the PopDownNumberDialog class and resources.
     *
     * When PopDownNumberPicker is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        PopDownDialog_1.PopDownDialog.init();
        if (registered === false) {
            window.customElements.define(PopDownNumberDialog.TAG_NAME.toLowerCase(), PopDownNumberDialog);
            registered = true;
        }
    }
    _initProperties() {
        this._titlePrimary = "";
        this._titleSecondary = "";
        this._laterHandle = null;
        this._extraCssFiles = [];
    }
    getValue() {
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        return textInput.valueAsNumber;
    }
    setValue(value) {
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        textInput.valueAsNumber = value;
    }
    setMinimum(min) {
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        textInput.setAttribute("min", "" + min);
    }
    setMaximum(max) {
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        textInput.setAttribute("max", "" + max);
    }
    setTitlePrimary(text) {
        this._titlePrimary = text;
        const dialog = DomUtils.getShadowId(this, ID_DIALOG);
        if (dialog != null) {
            dialog.setTitlePrimary(text);
        }
    }
    getTitlePrimary() {
        return this._titlePrimary;
    }
    setTitleSecondary(text) {
        this._titleSecondary = text;
        const dialog = DomUtils.getShadowId(this, ID_DIALOG);
        if (dialog != null) {
            dialog.setTitleSecondary(text);
        }
    }
    getTitleSecondary() {
        return this._titleSecondary;
    }
    /**
     * Specify extra Css files to load into this element.
     *
     * @param extraCssFiles extra Css files which should be loaded along side the default set.
     */
    addExtraCss(extraCssFiles) {
        this._extraCssFiles = [...this._extraCssFiles, ...extraCssFiles];
        this.updateThemeCss();
    }
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
        <${PopDownDialog_1.PopDownDialog.TAG_NAME} id="${ID_DIALOG}">
          <div class="form-group"><input type="number" id="${ID_INPUT}" min="0" max="10" class="form-control input-sm" value="1" /></div>
        </${PopDownDialog_1.PopDownDialog.TAG_NAME}>
        `;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.FONT_AWESOME, ...this._extraCssFiles];
    }
    //-----------------------------------------------------------------------
    /**
     *
     */
    handleKeyDown(ev) {
        if (ev.key === "Escape") {
            this._okId(null);
            ev.preventDefault();
            ev.stopPropagation();
            return;
        }
        if (ev.key === "Enter") {
            ev.preventDefault();
            ev.stopPropagation();
            const filterInput = DomUtils.getShadowId(this, ID_INPUT);
            this._okId(this.getValue());
        }
    }
    /**
     *
     */
    open(x, y, width, height) {
        const textInput = DomUtils.getShadowId(this, ID_INPUT);
        textInput.focus();
        const dialog = DomUtils.getShadowId(this, ID_DIALOG);
        dialog.open(x, y, width, height);
    }
    /**
     *
     */
    close() {
        const dialog = DomUtils.getShadowId(this, ID_DIALOG);
        dialog.close();
    }
    _okId(value) {
        if (this._laterHandle === null) {
            this._laterHandle = DomUtils.doLater(() => {
                this.close();
                this._laterHandle = null;
                const event = new CustomEvent("selected", { detail: { value: value } });
                this.dispatchEvent(event);
            });
        }
    }
}
/**
 * The HTML tag name of this element.
 */
PopDownNumberDialog.TAG_NAME = "ET-POPDOWNNUMBERDIALOG";
exports.PopDownNumberDialog = PopDownNumberDialog;
//# sourceMappingURL=PopDownNumberDialog.js.map