"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const Logger_1 = require("../Logger");
const ID = "EtScrollbarTemplate";
const ID_AREA = "ID_AREA";
const ID_CONTAINER = "ID_CONTAINER";
let registered = false;
/**
 * A scrollbar.
 */
class ScrollBar extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties(); // Initialise our properties. The constructor was not called.
    }
    /**
     * Initialize the Scrollbar class and resources.
     *
     * When Scrollbar is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(ScrollBar.TAG_NAME.toLowerCase(), ScrollBar);
            registered = true;
        }
    }
    _initProperties() {
        this._position = 0;
        this._length = 1;
        this._log = Logger_1.getLogger(ScrollBar.TAG_NAME, this);
    }
    /**
     * Custom Element 'connected' life cycle hook.
     */
    connectedCallback() {
        super.connectedCallback();
        if (DomUtils.getShadowRoot(this) === null) {
            // Set up the structure in the shadow DOM.
            const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
            const clone = this._createClone();
            shadow.appendChild(clone);
            this.updateThemeCss();
            this._getById(ID_CONTAINER).addEventListener('scroll', (ev) => {
                const container = this._getById(ID_CONTAINER);
                const top = container.scrollTop;
                this._position = top;
                // FIXME this should fire standard scroll events, not custom events.
                const event = new CustomEvent('scroll', { detail: {
                        position: top,
                        isTop: top === 0,
                        isBottom: (container.scrollHeight - container.clientHeight) === top
                    } });
                this.dispatchEvent(event);
            });
            this._updateLength(this.getAttribute(ScrollBar.ATTR_LENGTH));
            this._updatePosition(this.getAttribute(ScrollBar.ATTR_POSITION));
        }
        else {
            // Being reattached.
            this._updatePositionNumber(this._position);
        }
    }
    static get observedAttributes() {
        return [ScrollBar.ATTR_LENGTH, ScrollBar.ATTR_POSITION];
    }
    /**
     * Custom Element 'attribute changed' hook.
     */
    attributeChangedCallback(attrName, oldValue, newValue) {
        switch (attrName) {
            case ScrollBar.ATTR_LENGTH:
                this._updateLength(newValue);
                break;
            case ScrollBar.ATTR_POSITION:
                this._updatePosition(newValue);
                break;
            default:
                break;
        }
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_SCROLLBAR];
    }
    //-----------------------------------------------------------------------
    _createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
<div id='${ID_CONTAINER}'><div id='${ID_AREA}'></div></div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
    setLength(value) {
        if (value !== this._length) {
            this._length = Math.max(0, value);
            this._updateLengthNumber(this._length);
        }
    }
    getLength() {
        return this._getLength();
    }
    _getLength() {
        return this._length;
    }
    _updateLength(value) {
        if (value === null || value === undefined) {
            return;
        }
        const numberValue = parseInt(value, 10);
        if (isNaN(numberValue)) {
            console.warn("Value '" + value + "'to scrollbar attribute '" + ScrollBar.ATTR_LENGTH + "' was NaN.");
            return;
        }
        this.setLength(numberValue);
    }
    _updateLengthNumber(value) {
        const areaElement = this._getById(ID_AREA);
        areaElement.style.height = value + "px";
    }
    // --- Position attribute ---
    setPosition(value) {
        this._setPosition(value);
    }
    _setPosition(value) {
        const container = this._getById(ID_CONTAINER);
        const cleanValue = Math.min(container.scrollHeight - container.clientHeight, Math.max(0, value));
        if (cleanValue !== this._position) {
            this._position = cleanValue;
            this._updatePositionNumber(this._position);
        }
    }
    getPosition() {
        return this._getPosition();
    }
    _getPosition() {
        return this._position;
    }
    _updatePosition(value) {
        if (value === null || value === undefined) {
            return;
        }
        const numberValue = parseInt(value, 10);
        if (isNaN(numberValue)) {
            console.warn("Value '" + value + "'to scrollbar attribute '" + ScrollBar.ATTR_LENGTH + "' was NaN.");
            return;
        }
        this.setPosition(numberValue);
    }
    _updatePositionNumber(value) {
        const containerElement = this._getById(ID_CONTAINER);
        containerElement.scrollTop = value;
    }
    // --- Thumbsize attribute ---
    setThumbSize(size) {
    }
    getThumbSize() {
        return 7734; // FIXME bogus.
    }
    refresh(level) {
        this._updatePositionNumber(this._position);
    }
}
/**
 * The HTML tag name of this element.
 */
ScrollBar.TAG_NAME = 'ET-SCROLLBAR';
ScrollBar.ATTR_LENGTH = "length";
ScrollBar.ATTR_POSITION = "position";
ScrollBar.ATTR_THUMBSIZE = "thumbsize";
exports.ScrollBar = ScrollBar;
//# sourceMappingURL=ScrollBar.js.map