/**
 * Copyright 2016 Simon Edwards <simon@simonzone.com>
 */
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ViewerElement_1 = require("../ViewerElement");
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const ViewerElementTypes = require("../ViewerElementTypes");
const VirtualScrollArea = require("../VirtualScrollArea");
const Logger_1 = require("../Logger");
const VisualState = ViewerElementTypes.VisualState;
const ID = "EtImageViewerTemplate";
const ID_CONTAINER = "ID_CONTAINER";
const ID_CURSOR = "ID_CURSOR";
const ID_IMAGE = "ID_IMAGE";
const CLASS_FORCE_FOCUSED = "force-focused";
const CLASS_FORCE_UNFOCUSED = "force-unfocused";
const CLASS_FOCUS_AUTO = "focus-auto";
const SCROLL_STEP = 128;
const KEYBINDINGS_SELECTION_MODE = "image-viewer";
const COMMAND_GO_UP = "goUp";
const COMMAND_GO_DOWN = "goDown";
const DEBUG_SIZE = false;
let registered = false;
let instanceIdCounter = 0;
class ImageViewer extends ViewerElement_1.ViewerElement {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    static init() {
        if (registered === false) {
            window.customElements.define(ImageViewer.TAG_NAME.toLowerCase(), ImageViewer);
            registered = true;
        }
    }
    /**
     * Type guard for detecting a EtTerminalViewer instance.
     *
     * @param  node the node to test
     * @return      True if the node is a EtTerminalViewer.
     */
    static is(node) {
        return node !== null && node !== undefined && node instanceof ImageViewer;
    }
    _initProperties() {
        this._log = Logger_1.getLogger(ImageViewer.TAG_NAME, this);
        this._keyBindingManager = null;
        this._text = null;
        this._buffer = null;
        this._mimeType = null;
        this._imageWidth = -1;
        this._imageHeight = -1;
        this._cursorTop = 0;
        this._height = 0;
        this._mode = ViewerElementTypes.Mode.DEFAULT;
        this.document = document;
        this._visualState = VisualState.UNFOCUSED;
        this._currentElementHeight = -1;
        this._viewportHeight = -1;
    }
    //-----------------------------------------------------------------------
    //
    // ######                                
    // #     # #    # #####  #      #  ####  
    // #     # #    # #    # #      # #    # 
    // ######  #    # #####  #      # #      
    // #       #    # #    # #      # #      
    // #       #    # #    # #      # #    # 
    // #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    getTitle() {
        return "Text";
    }
    getAwesomeIcon() {
        return "file-image-o";
    }
    getSelectionText() {
        return null;
    }
    focus() {
        if (DomUtils.getShadowRoot(this) === null) {
            return;
        }
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        DomUtils.focusWithoutScroll(containerDiv);
    }
    hasFocus() {
        const hasFocus = this === DomUtils.getShadowRoot(this).activeElement; // FIXME
        return hasFocus;
    }
    setVisualState(newVisualState) {
        if (newVisualState !== this._visualState) {
            if (DomUtils.getShadowRoot(this) !== null) {
                this._applyVisualState(newVisualState);
            }
            this._visualState = newVisualState;
        }
    }
    getVisualState() {
        return this._visualState;
    }
    // getText(): string {
    //   return null;
    // }
    // 
    // setText(newText: string): void {
    //   if (this._codeMirror === null) {
    //     this._text = newText;
    //   } else {
    //     this._codeMirror.getDoc().setValue(newText);
    //   }
    // }
    setMimeType(mimeType) {
        this._mimeType = mimeType;
    }
    getMimeType() {
        return this._mimeType;
    }
    setBytes(buffer, mimeType) {
        this.setMimeType(mimeType);
        if (DomUtils.getShadowRoot(this) === null) {
            this._buffer = buffer;
        }
        else {
            this._setImage(buffer, mimeType);
        }
    }
    setMode(newMode) {
        this._mode = newMode;
    }
    getMode() {
        return this._mode;
    }
    setEditable(editable) {
        // this._editable = editable;
    }
    getEditable() {
        // return this._editable;
        return false;
    }
    // VirtualScrollable
    getHeight() {
        return this._height;
    }
    // VirtualScrollable
    setDimensionsAndScroll(setterState) {
        if (setterState.heightChanged || setterState.yOffsetChanged) {
            if (DEBUG_SIZE) {
                this._log.debug("setDimensionsAndScroll(): ", setterState.height, setterState.heightChanged, setterState.yOffset, setterState.yOffsetChanged);
            }
            this._adjustHeight(setterState.height);
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            if (containerDiv !== null) {
                containerDiv.scrollTop = setterState.yOffset;
            }
        }
    }
    // VirtualScrollable
    getMinHeight() {
        return 0;
    }
    // VirtualScrollable
    getVirtualHeight(containerHeight) {
        // const result = this.getVirtualTextHeight();
        let result = 0;
        if (this._imageHeight > 0) {
            result = this._imageHeight;
        }
        if (DEBUG_SIZE) {
            this._log.debug("getVirtualHeight: ", result);
        }
        return result;
    }
    // VirtualScrollable
    getReserveViewportHeight(containerHeight) {
        if (DEBUG_SIZE) {
            this._log.debug("getReserveViewportHeight: ", 0);
        }
        return 0;
    }
    // isFontLoaded(): boolean {
    //   return this._effectiveFontFamily().indexOf(NO_STYLE_HACK) === -1;
    // }
    getCursorPosition() {
        const detail = {
            left: 0,
            top: this._cursorTop,
            bottom: this._cursorTop + this._height,
            viewPortTop: this._cursorTop
        };
        return detail;
    }
    setCursorPositionTop(ch) {
        this._cursorTop = 0;
        this.focus();
        return true;
    }
    setCursorPositionBottom(ch) {
        this._cursorTop = this._imageHeight - this._height;
        this.focus();
        return true;
    }
    // From viewerelementtypes.SupportsMimeTypes
    static supportsMimeType(mimeType) {
        return ["image/x-bmp", "image/bmp", "image/png", "image/gif", "image/jpeg", "image/webp"].indexOf(mimeType) !== -1;
    }
    connectedCallback() {
        super.connectedCallback();
        if (DomUtils.getShadowRoot(this) !== null) {
            return;
        }
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        this.style.height = "0px";
        containerDiv.addEventListener('keydown', this._handleContainerKeyDown.bind(this));
        containerDiv.addEventListener('focus', (ev) => {
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            this._cursorTop = containerDiv.scrollTop;
        });
        const imgElement = DomUtils.getShadowId(this, ID_IMAGE);
        imgElement.addEventListener('load', this._handleImageLoad.bind(this));
        this._applyVisualState(this._visualState);
        if (this._buffer !== null) {
            this._setImage(this._buffer, this._mimeType);
        }
        this._adjustHeight(this._height);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.IMAGE_VIEWER];
    }
    //-----------------------------------------------------------------------
    //
    // ######                                      
    // #     # #####  # #    #   ##   ##### ###### 
    // #     # #    # # #    #  #  #    #   #      
    // ######  #    # # #    # #    #   #   #####  
    // #       #####  # #    # ######   #   #      
    // #       #   #  #  #  #  #    #   #   #      
    // #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}">
        </style>
        <div id="${ID_CONTAINER}" class="${CLASS_FORCE_UNFOCUSED}" tabindex="-1"><div id="${ID_CURSOR}"></div><img id="${ID_IMAGE}" /></div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _applyVisualState(visualState) {
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        containerDiv.classList.remove(CLASS_FORCE_FOCUSED);
        containerDiv.classList.remove(CLASS_FORCE_UNFOCUSED);
        containerDiv.classList.remove(CLASS_FOCUS_AUTO);
        switch (visualState) {
            case VisualState.AUTO:
                containerDiv.classList.add(CLASS_FOCUS_AUTO);
                break;
            case VisualState.FOCUSED:
                containerDiv.classList.add(CLASS_FORCE_FOCUSED);
                break;
            case VisualState.UNFOCUSED:
                containerDiv.classList.add(CLASS_FORCE_UNFOCUSED);
                break;
        }
    }
    _emitVirtualResizeEvent() {
        if (DEBUG_SIZE) {
            this._log.debug("_emitVirtualResizeEvent");
        }
        VirtualScrollArea.emitResizeEvent(this);
    }
    _emitBeforeSelectionChangeEvent() {
        const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_BEFORE_SELECTION_CHANGE, { bubbles: true });
        this.dispatchEvent(event);
    }
    _setImage(buffer, mimeType) {
        const dataUrl = DomUtils.CreateDataUrl(buffer, mimeType);
        const imageEl = DomUtils.getShadowId(this, ID_IMAGE);
        imageEl.setAttribute("src", dataUrl);
    }
    _handleImageLoad() {
        const imgElement = DomUtils.getShadowId(this, ID_IMAGE);
        this._imageWidth = imgElement.width;
        this._imageHeight = imgElement.height;
        const cursorDiv = DomUtils.getShadowId(this, ID_CURSOR);
        cursorDiv.style.height = "" + imgElement.height + "px";
        this._emitVirtualResizeEvent();
    }
    // ----------------------------------------------------------------------
    //
    //   #    #                                                 
    //   #   #  ###### #   # #####   ####    ##   #####  #####  
    //   #  #   #       # #  #    # #    #  #  #  #    # #    # 
    //   ###    #####    #   #####  #    # #    # #    # #    # 
    //   #  #   #        #   #    # #    # ###### #####  #    # 
    //   #   #  #        #   #    # #    # #    # #   #  #    # 
    //   #    # ######   #   #####   ####  #    # #    # #####  
    //                                                        
    // ----------------------------------------------------------------------
    dispatchEvent(ev) {
        if (ev.type === 'keydown' || ev.type === 'keypress') {
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            return containerDiv.dispatchEvent(ev);
        }
        else {
            return super.dispatchEvent(ev);
        }
    }
    _handleContainerKeyDown(ev) {
        if (this._keyBindingManager !== null && this._keyBindingManager.getKeyBindingContexts() !== null &&
            this._mode === ViewerElementTypes.Mode.CURSOR) {
            const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_SELECTION_MODE);
            if (keyBindings !== null) {
                const command = keyBindings.mapEventToCommand(ev);
                if (command !== null) {
                    const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
                    ev.preventDefault();
                    ev.stopPropagation();
                    switch (command) {
                        case COMMAND_GO_UP:
                            // Cursor up
                            if (this._cursorTop !== 0) {
                                const newTop = Math.max(this._cursorTop - SCROLL_STEP, 0);
                                this._cursorTop = newTop;
                                const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE, { bubbles: true });
                                this.dispatchEvent(event);
                            }
                            else {
                                const detail = { edge: ViewerElementTypes.Edge.TOP, ch: 0 };
                                const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_EDGE, { bubbles: true, detail: detail });
                                this.dispatchEvent(event);
                            }
                            break;
                        case COMMAND_GO_DOWN:
                            // Cursor down          
                            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
                            if (this._cursorTop + this._height < this._imageHeight) {
                                const newTop = Math.min(this._imageHeight - this._height, this._cursorTop + SCROLL_STEP);
                                this._cursorTop = newTop;
                                const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE, { bubbles: true });
                                this.dispatchEvent(event);
                            }
                            else {
                                const detail = { edge: ViewerElementTypes.Edge.BOTTOM, ch: 0 };
                                const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_EDGE, { bubbles: true, detail: detail });
                                this.dispatchEvent(event);
                            }
                            break;
                        default:
                            break;
                    }
                    ev.preventDefault();
                    ev.stopPropagation();
                    return;
                }
            }
        }
    }
    _getClientYScrollRange() {
        return Math.max(0, this.getVirtualHeight(this.getHeight()) - this.getHeight() + this.getReserveViewportHeight(this.getHeight()));
    }
    _adjustHeight(newHeight) {
        this._height = newHeight;
        if (this.parentNode === null || DomUtils.getShadowRoot(this) === null) {
            return;
        }
        if (newHeight !== this._currentElementHeight) {
            this._currentElementHeight = newHeight;
            this.style.height = "" + newHeight + "px";
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            containerDiv.style.height = "" + newHeight + "px";
        }
    }
}
ImageViewer.TAG_NAME = "ET-IMAGE-VIEWER";
exports.ImageViewer = ImageViewer;
function px(value) {
    if (value === null || value === undefined || value === "") {
        return 0;
    }
    return parseInt(value.slice(0, -2), 10);
}
//# sourceMappingURL=ImageViewer.js.map